;  File: LIFE.ASM      System: Simulation    Version: 1.11  Date: 04-15-95  ;

;-----------------------------------------------------------------------------
; LIFE is cellular automata "game" from Martin Gardner column in Scientific
; American magazine, October, 1970.  Initial seeding of two-dimensional grid
; plus three simple life/death/birth rules determines evolution.
;
; Assemble with MASM LIFE /DDOCLEN n, where n is size of DOC file appended to
; COM file afterwards.  Minus documentation, COM file is 4K code and 1K data.
; Code is fairly tight to keep it under 4K.                       --C. Hessel
;
; 15 Apr 94 -- added second parm Q option to suppress beeps
;-----------------------------------------------------------------------------
Code_Seg   SEGMENT
           ASSUME CS:Code_Seg,DS:Code_Seg,ES:Code_Seg

MOVE       EQU xchg            ; Saves byte on some AX register-register moves

IWID       = 64                ; Initial square grid dimensions
SWID       = 60                ; Byte width/length of display window
MINWID     = 64                ; Lower cap on grid dimensions
MAXWID     = 720               ; Upper cap on grid dimensions
BITWID     = 8*SWID            ; Pixel width/length of display window
MAXPRN     = 592               ; Maximum grid dimension for print

MAXSIZE    = MAXWID/8*MAXWID   ; Maximum compressed grid bytes (just under 64K)
EDGEINC    = 8                 ; Grid edge expand/reduce increment (8-multiple)
CORNERLEN  = 8                 ; Zoom box corner length

PANELWID   = 19                ; Width of help panel text
DOTWID     = 600               ; HP 75 dpi width/length of 8.5x11 sheet
DOTLEN     = 808               ;   allowing for HP unprintable border strips
HISTLEN    = 128               ; Byte length of history buffer--two power

EDGEALARM  = 1                 ; Edge alarm--bit flags in Flags byte
HISTALARM  = 2                 ; Cycle alarm
RCRDFLAG   = 4                 ; Macro recording flag
PLAYFLAG   = 8                 ; Macro playing flag
CLRNUMFLAG = 16                ; Flag to clear CurNum
INNUMFLAG  = 32                ; Flag to accumulate CurNum
RESUMEFLAG = 64                ; Free-running evolution flag
HIRESFLAG  = 128               ; High resolution display flag

CSRREPMAX  = MAXWID            ; Cap on Cursor/Toggle repetitions
GENREPMAX  = 1000              ; Cap on Next Step repetitions

MACREADMAX = 8000h             ; Macro read buffer size (before strip)
MACMAX     = 4096              ; Macro buffer sizes
MACNEST    = 16                ; Macro nesting limit
MSTKSIZE   = 4*MACNEST         ; Size of macro state stack

BRDRCOLR   = 0                 ; Black
BKGNDCOLR  = 1                 ; Blue
HELPCOLR   = 14                ; Yellow
FORECOLR   = 15                ; White
BITCOLR1   = 14                ; Yellow
BITCOLR2   = 12                ; Light red

BITMODE    = 18                ; VGA 640x480, 16 color
TXTMODE    = 3                 ; 80x25 text, 16 color
BITSEG     = 0A000h            ; Graphics video segment
TXTSEG     = 0B800h            ; Text video segment

F1KEY      = 3Bh               ; F1 extended key code
UPKEY      = 48h               ; Cursor extended key codes
DNKEY      = 50h
LFKEY      = 4Bh
RTKEY      = 4Dh
PGDN       = 51h
PGUP       = 49h
HOMEKEY    = 47h
ENDKEY     = 4Fh

MSGRC      = 25 SHL 8 OR SWID+1      ; Help panel row/column locations
ALARMRC    = 27 SHL 8 OR SWID+1
FOOTRC     = 29 SHL 8 OR SWID+1
CENTERRC   = 14 SHL 8 OR (SWID/2-2)  ; Center location for Wait and Key?

TMR2PORT   = 42h               ; Beep routine ports
CTRLPORT   = 43h
SPKRPORT   = 61h
SPKRCTRL   = 00000011b         ; 1=enable speaker data/1=timer2 gate
TMR2CTRL   = 10110110b         ; 10=timer2/11=read low then high/
                               ;   011=square wave/0=binary
ZeroSeg    SEGMENT AT 00h      ; BIOS data references
           ORG   46Ch
Timer      DW    ?             ; Timer word for beep ticks/random seed
ZeroSeg    ENDS

           ORG   02h
EndSeg     LABEL WORD          ; Segment at end of COM allocation
           ORG   5Dh
CmdArg     LABEL BYTE          ; "?" or macro number
           ORG   6Dh
CmdArg2    LABEL BYTE          ; "Q" to suppress beeps
;-----------------------------------------------------------------------------
; Initialization code.
;-----------------------------------------------------------------------------
           ORG   100h          ; COM file start
Start:     mov   dx,OFFSET NewInt24h
           mov   ax,2524h      ; Install new critical error handler,
           int   21h           ;   mainly for open drive door

           cld                 ; Program default
           xor   cx,cx         ; File attributes or later use
           cmp   CmdArg,"?"    ; Test if documentation dump wanted
           jne   notdump       ; No?  Ahead

           mov   ah,3Ch        ; Create--next points DX to filename
           call  HelpOpen      ; May beep abort internally
           call  Abort         ; Display message, beginning with filename
           mov   dx,OFFSET DocText
           mov   cx,DOCLEN     ; Equate passed on MASM command-line
           jmp   HelpWrite     ; Write, close, and exit via PSP

notdump:   mov   ax,1A00h      ; BIOS get display type
           int   10h
           cmp   al,1Ah        ; Test if function supported
           mov   dx,OFFSET MsgBuf
           jne   Abort         ; No?  Abort

           cmp   bl,8          ; Check for color VGA
           jne   Abort         ; No?  Abort

           mov   ax,ds
           add   ax,3000h      ; Ahead 192K
           cmp   ax,EndSeg     ; Check if enough memory
           jbe   chkok         ; Yes?  Ahead

Abort:     mov   ah,9          ; MsgBuf is error message initially
           int   21h
           ret                 ; Exit via PSP or return

chkok:     sub   ax,1000h      ; Back up 64K--all early exit checks done
           mov   TempSeg,ax    ; Fix as temp use segment
           sub   ax,1000h      ; Back up 64K
           mov   GridSeg,ax    ; Fix as grid segment

           call  OneTick       ; Returns low tick word AX
           mov   LoSeed,ax     ; Initialize random number seed, low word

           mov   di,dx         ; DX still MsgBuf and CX zero
           mov   cl,OFFSET EndLogBuf - OFFSET MsgBuf - 2
           mov   al," "
           rep   stosb         ; Clear MsgBuf/AlarmFoot/PanelFoot
           call  KATail        ; Stores CrLf afterward

           mov   ax,BITMODE    ; Set graphics mode til final exit
           int   10h           ; Set mode
           mov   ax,0B00h
           mov   bl,BKGNDCOLR  ; Blue background
           int   10h
           mov   ax,1001h
           mov   bh,BRDRCOLR   ; Black border
           int   10h

           mov   es,VidSeg     ; Draw vertical help panel border
           mov   di,SWID
           mov   cx,BITWID     ; Rows
           mov   al,10110100b
barloop:   stosb               ; Draw vertical panel bar
           add   di,79         ; Screen width 80
           loop  barloop

           call  ClrGrid       ; Clear grid, set Gen through Row, reset ES

           mov   di,OFFSET SumTbl
           xor   ax,ax         ; AH runs 0-255 and AL accumulates sum
mkoutlp:   push  ax            ; Save AH and (zero) AL
           mov   cl,8          ; CH zero from ClrGrid first time
mkinlp:    shr   ah,1
           adc   al,0          ; Accumulate sum of 8 bits
           loop  mkinlp

           stosb               ; Store sum 0-8 for current AH
           pop   ax            ; Restore AH and (zero) AL
           inc   ah
           jne   mkoutlp       ; At end, DI points to PlayBuf

           mov   si,OFFSET InitMac
           mov   cl,OFFSET PanelTitle - OFFSET InitMac
           rep   movsb         ; Copy macro, pointing SI to help panel

           mov   cl,30         ; Number of screen rows--CH zero
helploop:  push  cx            ; Loop displays help panel
           mov   ax,1E20h      ; 30 high, space low
           sub   ah,cl         ; Set AH to row
           xchg  al,[si]       ; Set AL to indent/length, blanking [SI]
           inc   si
           push  ax            ; Save for width
           mov   cl,5
           shr   al,cl
           add   al,SWID+1     ; Column
           MOVE  dx,ax         ; Row/col to DX
           pop   cx
           and   cl,31         ; Width
           call  DispHlpLn     ; Display one line
           pop   cx
           loop  helploop

           call  Alarm         ; Toggle alarms off (clearing alarm flags)
           call  RcrdOff       ; Display record off

           mov   si,OFFSET CmdArg
           call  IsDigitSI     ; Load AL and set carry if digit
           jnc   tojmp         ; No user macro specified?  Use demo macro

hnloop:    call  InitNum       ; Accumulate AL to CurNum
           call  IsDigitSI
           jc    hnloop        ; Eventual non-digit at 2nd argument drive

           call  FirstMac      ; Load macro--if not found, demo macro used
tojmp:     jmp   Pause         ; To main loop...
;-----------------------------------------------------------------------------
; Variable data.  See also EOF for buffers, etc.  Text "extracted" contained
; in CurFunc through PrnCol for convenience.  Initial values irrelevant for
; those variables, except that HiSeed must be non-zero.  Order of variables
; here significant for most part.  E.g., Gen through Row referenced as block
; in Reset, CellMax through GldCnt referenced as block in DispGen, and some
; words referenced as pairs.
;-----------------------------------------------------------------------------
Flags      DB    PLAYFLAG OR HISTALARM OR EDGEALARM
           EVEN
FileName   DB    "LIFE   .DOC",0     ; Digits/extension adjusted as needed
CurFunc    DW    "xe"                ; PopFunc parameter
CsrBump    DW    "rt"                ; Low/high is row/col increment
LoSeed     DW    "ca"                ; Low random fill control
HiSeed     DW    "et"                ; Insure non-zero high word initially
PrnCol     DW    "$d"                ; Upper left corner in absolute dots
PrnRow     DW    ?                   ;   (75 dpi) of last print grid
PrnWid     DW    0                   ; Last print grid size--zero is flag

CellMax    DW    0                   ; Maximum cell count
Wid        DW    IWID                ; Current grid edge size
CurNum     DW    0                   ; Accumulates prefix numbers
Gen        DW    ?                   ; Counts generation steps
GldCnt     DW    ?                   ; Escaped glider count
Delta      DW    ?                   ; Tracks cursor direction
Left       DW    ?                   ; Origin of small window
Top        DW    ?
Col        DW    ?                   ; Cursor position relative to previous
Row        DW    ?

LogName    DB    "LIFE.LOG"          ; Log file name, including next zero
GridOff    DW    0                   ; Grid offset fixed at zero
GridSeg    DW    ?                   ; Segment set at run time
TempOff    DW    0                   ; Temp use offset fixed at zero
TempSeg    DW    ?                   ; Segment set at run time
VidOff     DW    0                   ; Graphics video offset/segment
VidSeg     DW    BITSEG

MStkPtr    DW    OFFSET MStkBuf      ; Macro stack pointer
PlayNum    DW    0                   ; Number of playing macro
PlayPtr    DW    OFFSET PlayBuf      ; Pointer into macro play buffer
RcrdPtr    DW    ?                   ; Pointer into macro record buffer
;-----------------------------------------------------------------------------
; Fixed data.  See EOF for demo macro/help panel text/documentation.
;-----------------------------------------------------------------------------
BlkAdj     DB    1,1                 ; Block Delta's, includes start of next
BitAdj     DB    2,1,2,2,1,2,1,1     ; BitCsr Delta's to subtract

ByteCsr    DW    0011111011100011b   ; Left pointing
BitCsr     DW    0011111000110000b   ; Left pointing

RowList    DW    255  SHL 8 OR 255   ; List for zoom box corner draw
           DW    255  SHL 8 OR SWID+1-CORNERLEN
           DW    SWID SHL 8 OR 255
           DW    SWID SHL 8 OR SWID+1-CORNERLEN

Glider1    DW    1100101010000000b,0000h
           DW    0                         ; End List
Glider2    DW    0110110000100000b,0000h
           DW    0                         ; End List
Glider3    DW    1110100001000000b,0000h
           DW    0                         ; End List
Glider4    DW    0100110010100000b,0000h
           DW    0                         ; End List
Ship1      DW    0100100010001111b,0001h
           DW    0000000000101100b,0201h
           DW    0                         ; End List
Ship2      DW    0100100010001111b,0001h
           DW    0000000000101100b,0301h
           DW    0                         ; End List
Ship3      DW    0100100010001111b,0001h
           DW    0000000000101100b,0401h
           DW    0                         ; End List
Gun        DW    1100101010000000b,0000h   ; Glider 1
           DW    0011001100000000b,010Eh   ; Block
           DW    0111011100000000b,0208h   ; 2x3 block
           DW    0010000000110000b,0406h   ; 3 cells
           DW    0100000011000000b,0106h   ; 3 cells inverted
           DW    0011010010001000b,0FFF6h  ; Curve
           DW    1100001000010001b,02F6h   ; Inverted curve
           DW    0011001100000000b,0FFECh  ; Block
           DW    0                         ; End list

ObjTbl     DW   OFFSET Glider1,OFFSET Glider2,OFFSET Glider3,OFFSET Glider4
           DW   OFFSET Ship1,  OFFSET Ship2,  OFFSET Ship3,  OFFSET Gun
NUMOBJ     =    ($ - OFFSET ObjTbl) / 2

JmpTbl     DW    OFFSET Dummy,  OFFSET ClrGrid, OFFSET Veer,   OFFSET MacSave
           DW    OFFSET MacPlay,OFFSET MacRcrd, OFFSET Save,   OFFSET Load
           DW    OFFSET Eject,  OFFSET PrnGrid, OFFSET Zoom,   OFFSET Alarm
           DW    OFFSET KillChn,OFFSET RandFill,OFFSET Insert, OFFSET LogState
           DW    OFFSET Home,   OFFSET Reduce,  OFFSET Expand, OFFSET NextGen
           DW    OFFSET Toggle, OFFSET Jump,    OFFSET Drag

Letters    DB    "DJTNXRHGIFKAZPELSOMWVC",0; Inverse order of JmpTbl
NUMLET     =     $ - OFFSET Letters
NUMREPLET  =     6                         ; Left few are repeating controls

MacKeys    DB    "udlrpnhe",0              ; Inverse order of CsrKeys--includes
                                           ;   prior 0
CsrKeys    DB    ENDKEY,HOMEKEY,PGDN,PGUP,RTKEY,LFKEY,DNKEY,UPKEY
NUMCSR     =     $ - OFFSET CsrKeys + 1    ; Inverse order of Delta lists--
                                           ;   includes next 0
DeltaCol   DB    0,0                       ; Includes start of next
DeltaRow   DB    0,-1,1,0,0,-EDGEINC,EDGEINC,0,0

KillList   DB    01001011b, 01011010b, 10100101b, 10110100b
           DB    01101001b, 00101101b, 11010010b, 10010110b

AlarmTbl   DB    "Off  Edge CycleBoth "    ; Alarm settings
FullText   DB    "Full"                    ; For full macro-record buffer
WaitText   DB    "Wait"                    ; Fill/print/etc. message
KeyText    DB    "Key?"                    ; Macro input prompt
;-----------------------------------------------------------------------------
; Chain message and HP printer strings.  Blocks of ASCII 0's are overwritten
; as needed.  Note that HP_Pos and HP_Lines use 300 dpi resolution, independent
; of selected 75 dpi raster resolution.  Also, with no top margin specified
; below, about 1/6" unprintable vertical space must be skipped in PrnGrid.
;-----------------------------------------------------------------------------
ChainMsg   DB    "Chain count/length:  "   ; For extended log data
           DB    "00000/000 ",13,10        ; Digits adjusted, possible + sign
CHNMSGLEN  =     $ - OFFSET ChainMsg

HP_Reset   DB    27,"E"                    ; Reset

HP_Init    DB    27,"&loE"                 ; Portrait and no top margin
           DB    27,"9"                    ; Clear horizontal margins
           DB    27,"(sp16.66h8.5vsbT"     ; Line printer 16.66 cpi (internal)
INITLEN    =     $ - OFFSET HP_Init

HP_Stop    DB    27,"*rB"                  ; End raster
HP_Pos     DB    27,"*p0000x0000Y"         ; Dot position (300 dpi)
STOPLEN    =     $ - OFFSET HP_Stop        ; For both raster stop and position

HP_Lines   DB    27,"*c0000a1bp0000b1aP"   ; Horizontal and vertical lines
POSLEN     =     $ - OFFSET HP_Pos         ; For both position and line draw

HP_Rast:   DB    27,"*t75R"                ; 75 DPI raster
           DB    27,"*bM"                  ; Unencoded
           DB    27,"*r1A"                 ; Start at cursor
RASTLEN    =     $ - OFFSET HP_Rast

HP_Xfer    DB    27,"*b0000W"              ; Transfer # bytes
XFERLEN    =     $ - OFFSET HP_Xfer
;-----------------------------------------------------------------------------
; Critical error handler, basically for open-drive door on file I/O.
;-----------------------------------------------------------------------------
NewInt24h: mov   al,0          ; Flag abort
           iret
;-----------------------------------------------------------------------------
; Sets carry if AL is digit, else clears carry.
;-----------------------------------------------------------------------------
IsDigitSI: lodsb
IsDigit:   cmp   al,"0"
           cmc
           jnc   idout

           cmp   al,"9"+1
idout:     ret
;-----------------------------------------------------------------------------
; Main loop is Pause/Resume coroutine, with program exit internal to GetKey.
;-----------------------------------------------------------------------------
doedit:    call  Edit
           call  ChkClr        ; Clear CurNum, etc., if flagged
Pause:     and   Flags,NOT RESUMEFLAG
           call  DspGenCsr
psreget:   call  GetKey
           jc    psreget       ; No key?  Continue looking
           jne   doedit        ; Not switch key?  Handle edit key

Resume:    or    Flags,RESUMEFLAG
           call  RcrdOff
           call  PlayOff       ; In case F1 at macro input
rsnext:    call  NextGen1
rsdisp:    call  DispGen
           call  ChkAlarms     ; Check for edge/cycle alarm
           jc    Pause         ; Yes?  Pause

           call  GetKey
           jc    rsnext        ; No key?  Continue evolution
           je    Pause         ; Switch key?  To edit state

           call  GetCsr        ; Get AL/AH cursor increment, possibly zero
           mov   di,OFFSET Left
           call  BumpPair      ; Bump Left/Top by AL/AH
           jmp   SHORT rsdisp  ; Look for repeated keypress before resuming
;-----------------------------------------------------------------------------
; Return next macro key AX or set carry if no macro playing.  Macro-specific
; functions are handled here also.
;-----------------------------------------------------------------------------
GetMacKey: test  Flags,PLAYFLAG
           je    stcout        ; No macro running?  Out with set carry

gmkloop:   call  DispPlay      ; Display play count
           mov   si,PlayPtr
           lodsb               ; Fetch character
           mov   PlayPtr,si    ; Update pointer
           and   ax,0FFh       ; Zero AH to anticipate not lower case
           je    chkpop        ; End?  Check for prior macro

           cmp   al,"!"
           jne   chkinp        ; Not macro Esc equivalent?  Ahead

           mov   al,27         ; Else convert to Esc
           ret                 ; Exit with clear carry

chkinp:    cmp   al,"?"
           jne   chkqt         ; Not input prompt?  Ahead

           mov   si,OFFSET KeyText
           call  DispChr4      ; Display prompt
BiosKey:   mov   ah,0          ; Bios get key, with carry cleared
           int   16h
gmkout:    clc
           ret

chkqt:     cmp   al,'"'
           jne   chklow        ; Not message literal?  Ahead

           call  ShowMsg       ; Extract/display message, advancing PlayPtr
           jmp   SHORT gmkloop ; Loop to get next macro character

chklow:    cmp   al,"a"        ; Check for lower case, flagging extended key
           jb    gmkout        ; Not extended key?  Out

           mov   cx,NUMCSR     ; List includes trailing dummy zero
           mov   di,OFFSET MacKeys
           repne scasb         ; Clears carry, even if not equal
           mov   di,cx
           mov   ax,WORD PTR CsrKeys[di-2]
           mov   al,0          ; Got AH, but flag extended key
           ret

chkpop:    call  PopMac        ; Restore prior macro, if any
           jnc   gmkloop       ; Prior macro?  Loop, else stop play
;-----------------------------------------------------------------------------
; Terminate playing macro, if any.  Set carry on exit.
;-----------------------------------------------------------------------------
StopPlay:  test  Flags,PLAYFLAG
           je    stcout        ; No macro running?  Avoid screen display

           call  DspGenCsr     ; In case Wait message still around
PlayOff:   call  ClrMsg        ; Clear message display
           call  ClrNum        ; Clear possible pending number
           call  SetMStack     ; Insure stack pointer reset
           mov   di,OFFSET AlarmFoot+6
           mov   al,NOT PLAYFLAG
           call  FlagOff       ; Turn off flag and display Off
stcout:    stc
           ret
;-----------------------------------------------------------------------------
; Get key AX, with internal abort on Esc.  Set carry flags no key.  Clear
; carry and equality flags switch key.  Clear carry and inequality flags other.
;-----------------------------------------------------------------------------
GetKey:    mov   ah,1          ; Check if key waiting--priority over macro
           int   16h
           jne   gkget         ; Key?  Ahead

           call  GetMacKey     ; Else get next macro key
           jc    gkout         ; No macro running?  Out with set carry
           jmp   SHORT gkup

gkget:     call  StopPlay      ; Any true key aborts macros
           call  BiosKey       ; Fetch key
gkup:      cmp   al,27
           je    done          ; Esc?  End program

           cmp   ax,F1KEY SHL 8
           clc                 ; Else exit with clear carry and
gkout:     ret                 ;   equality/inequality for switch key

done:      pop   ax            ; Kill return address
           mov   ax,TXTMODE
           int   10h           ; Set mode back to color text
           mov   ax,TXTSEG
           mov   es,ax
           xor   di,di         ; Clear screen
           mov   cx,25*80
           mov   ax,7 SHL 8 OR " "
           rep   stosw
           ret                 ; PSP exit--no need to restore ES
;-----------------------------------------------------------------------------
; Get cursor increment AH/AL from extended key code AH.  If good key, also
; set carry and adjust Delta.
;-----------------------------------------------------------------------------
GetCsr:    mov   al,ah         ; Ok on Pan if AL not zero
           mov   cx,NUMCSR     ; Last element 0 translates to no move
           mov   di,OFFSET CsrKeys
           repne scasb         ; Scan list to get CX
           mov   di,cx
           mov   ax,WORD PTR DeltaRow[di-1]
           mov   al,DeltaCol[di]
           test  ax,ax
           je    gcout         ; Got dummy zero?  No Delta change

           push  ax
           call  SetDelta      ; Adjust Delta, preserving DI
           pop   ax
           stc                 ; Good flag for edit call (and DI used too)
gcout:     ret
;-----------------------------------------------------------------------------
; Convert row/column increments AH/AL to direction AX and save as Delta.
;-----------------------------------------------------------------------------
SetDelta:  call  SwapPMZ       ; Preserves DI
SwapPMZ:   xchg  al,ah
           test  al,al         ; Convert AL to +1/-1/0
           je    sdsave

           mov   al,1
           jns   sdsave

           neg   al
sdsave:    mov   Delta,ax      ; Second save sticks
           ret
;-----------------------------------------------------------------------------
; Veer right (direction change) by adjusting Delta.
;-----------------------------------------------------------------------------
Veer:      mov   ax,0FFh
           call  AdjAHAL       ; Adjust according to current direction
           jmp   SHORT sdsave  ; Save new Delta
;-----------------------------------------------------------------------------
; Drag cell chain in cursor direction.
;-----------------------------------------------------------------------------
Drag:      mov   cx,CSRREPMAX  ; Cap repetitions
           call  PopFunc       ; Careful--pops next address
Drag1:     mov   di,4          ; Point near start of temp segment
           push  di
           call  KillDrag      ; Kill chain, saving kill history
           pop   si            ; Start of kill history
           test  bp,bp         ; Kill Count
           je    dragout       ; None?  Ahead for cursor move

dragloop:  mov   ax,Delta      ; Loop through kill history, drawing
           mov   ds,TempSeg    ;   at locations bumped by Delta
           mov   di,si
           call  BumpPair      ; Bump DS:[DI]/[DI+2] by AL/AH sign extended
           lodsw
           MOVE  cx,ax         ; Adjusted row
           lodsw
           MOVE  dx,ax         ; Adjusted column
           push  cs
           pop   ds

           call  SetCell       ; Preserves DX/CX/BP/SI/DI
           dec   bp
           jne   dragloop

dragout:   jmp   SHORT deltmove; Advance in cursor direction
;-----------------------------------------------------------------------------
; Toggle cell, with cursor movement afterward in cursor direction.
;-----------------------------------------------------------------------------
Toggle:    mov   cx,CSRREPMAX  ; Cap on repetitions
           call  PopFunc       ; Careful--pops next address
Toggle1:   call  SetDXCX
           call  GetCell
           cmc
           call  PutCell
deltmove:  mov   ax,Delta      ; Advance cursor
           jmp   SHORT BumpRwCl
;-----------------------------------------------------------------------------
; Move cursor in current direction.  Bump... entry points adjust either
; Col/Row or Left/Top at DI by increments AL/AH, with AX preserved.
;-----------------------------------------------------------------------------
Jump:      mov   ax,Delta
           mov   CsrBump,ax
CursMove:  mov   cx,CSRREPMAX  ; Cap on repetitions
           call  PopFunc       ; Careful--pops next address
CursMove1: mov   ax,CsrBump
BumpRwCl:  mov   di,OFFSET Col ; Entry from Toggle
BumpPair:  mov   dx,ax         ; Entry from Resume
           cbw                 ; Sign extend, since negative allowed
           add   [di],ax
           mov   al,dh
           cbw                 ; Ditto
           add   [di+2],ax
           MOVE  ax,dx         ; AX preserved on Bump... calls
           ret
;-----------------------------------------------------------------------------
; Handle edit key in AX.
;-----------------------------------------------------------------------------
Edit:      cmp   al,0
           jne   chknum        ; Not extended key?  Check digit

           or    Flags,CLRNUMFLAG
           call  GetCsr        ; Set AL/AH row/col increments, Delta/DI/carry
           mov   CsrBump,ax    ; Save for possible CursMove
           pushf               ; Set carry means good key
           mov   al,MacKeys[di-1]
           call  ChkRcrd       ; Record key (or dummy dot)
           popf
           jc    CursMove      ; Good key?  Back

           ret

chknum:    call  IsDigit       ; Set carry if AL is digit
           jnc   chklet        ; Not digit?  Check for letter

           push  ax            ; Save digit
           call  ChkRcrd       ; Record, as needed
           pop   ax
           or    Flags,INNUMFLAG
InitNum:   and   ax,0Fh        ; To binary 0-9
           MOVE  cx,ax         ; Save
           mov   ax,10
           mul   CurNum
           add   ax,cx         ; Plus input
           mov   CurNum,ax     ; Update number
           ret

chklet:    or    Flags,CLRNUMFLAG
           and   al,NOT "a"-"A"; Force upper case
           mov   cx,NUMLET     ; Includes trailing dummy space
           mov   di,OFFSET Letters
           repne scasb
           push  cx            ; Save list index
           je    goodlet       ; Found good key or 0?  Ahead

           mov   al,0          ; For ChkRcrd only
goodlet:   call  ChkRcrd
           pop   di            ; Restore list index
           shl   di,1          ; Double for word size of JmpTbl
           jmp   JmpTbl[di]    ; To appropriate routine...
;-----------------------------------------------------------------------------
; Clear CurNum, if flagged, with associated flags.
;-----------------------------------------------------------------------------
ChkClr:    test  Flags,CLRNUMFLAG
           je    Dummy         ; Clear not wanted?  Out

ClrNum:    xor   ax,ax
           mov   CurNum,ax
           and   Flags,NOT (INNUMFLAG OR CLRNUMFLAG)
Dummy:     ret
;-----------------------------------------------------------------------------
; Execute function on stack CurNum times with CX repetition cap, if flagged,
; else just once.  Break if key waiting.  Insure called functions have no
; register input.
;-----------------------------------------------------------------------------
PopFunc:   pop   ax            ; Careful--call must precede expected function
           mov   CurFunc,ax    ; Store for loop calls
           call  CapNum        ; Set AX to repetiton count
           test  Flags,INNUMFLAG
           jne   usenum        ; Repetitons wanted?  Ahead

           inc   ax            ; Else do 1 call--CurNum zero if not in number
usenum:    mov   CurNum,ax     ; Loop counter now
           test  ax,ax         ; Zero only if explicit, e.g., 0T or 0N
           jne   rfent         ; Call function, bypassing initial display
                               ;   and leaving final display for after exit
           ret                 ; Explicit zero prefix?  Do nothing

reploop:   mov   ah,1
           int   16h
           jne   repfout       ; Key waiting?  Early out

           call  ChkBoth       ; Sets inequality if operation overshot grid
           jne   repfout       ; Yes?  Exit Toggle/Jump/Cursor/Drag,
                               ;   but Reduce washed out earlier
           call  DspGenCsr
           cmp   CurFunc,OFFSET NextGen1
           jne   rfent         ; Not Next repetition?  Omit alarm check

           call  ChkAlarms     ; Check for edge/cycle alarm
           jc    repfout       ; Alarm?  Early out

rfent:     call  [CurFunc]     ; Call function
           dec   CurNum
           jne   reploop       ; Loop til zero

repfout:   ret
;-----------------------------------------------------------------------------
; Return minimum of CurNum and CX in AX.  CX preserved.
;-----------------------------------------------------------------------------
CapNum:    mov   ax,CurNum
           cmp   ax,cx
           jbe   cnout

           mov   ax,cx
cnout:     ret
;-----------------------------------------------------------------------------
; Insert object specified by CurNum into grid.
;-----------------------------------------------------------------------------
Insert:    mov   cx,NUMOBJ-1
           call  CapNum
           shl   ax,1          ; Double for word offset table
           MOVE  si,ax
           mov   si,ObjTbl[si] ; List start
           jmp   SHORT PutList
;-----------------------------------------------------------------------------
; Adjust increments AH/AL according to current cursor direction.  Called
; by PutList and Veer.
;-----------------------------------------------------------------------------
AdjAHAL:   MOVE  cx,ax
           call  GetDirec      ; Set AX to 0-3
           inc   ax
           xchg  cx,ax         ; CX to 1-4 and restore AX
aaloop:    xchg  ah,al         ; Else convert to AH/AL to row/col
           neg   al            ;   increments based on direction
           loop  aaloop

           ret
;-----------------------------------------------------------------------------
; Draw list of 4x4 blocks at SI.
;-----------------------------------------------------------------------------
plloop:    MOVE  bp,ax         ; 16-bit image
           lodsw               ; Get adjustment
           call  AdjAHAL       ; Adjust according to current direction
           MOVE  cx,ax         ; Next call forces back to AX
           call  SetDXCX       ; Set cursor DX/CX, old CX to AX
           call  AdjDXCX       ; Adjust DX/CX from AH/AL--below
           push  si
           mov   si,OFFSET BlkAdj
           call  PrepBlk       ; Rotate BP and adjust DX/CX--below
           mov   si,OFFSET PutCell
           call  DoSI4x4       ; Insert 4x4 BP at DX/CX
           pop   si
PutList:   lodsw               ; Get next 4x4 block
           test  ax,ax
           jne   plloop        ; Not list end?  Loop

           ret
;-----------------------------------------------------------------------------
; From direction 0-3 in AX, position DX/CX, 4x4 image BP, and table at SI,
; rotate BP and adjust DX/CX.  Called by Insert and DispCsr.
;-----------------------------------------------------------------------------
PrepBlk:   call  RotBlk        ; Rotate BP, with DX/CX/SI preserved

           call  GetDirec      ; Set AX to 0-3 for direction
           add   si,ax
           add   si,ax
           lodsw               ; Get AH/AL adjustmenst for DX/CX

AdjDXCX:   push  ax            ; Save AH
           cbw
           sub   cx,ax         ; Adjust column CX
           pop   ax            ; Restore AH
           mov   al,ah
           cbw
           sub   dx,ax         ; Adjust row DX
           ret
;-----------------------------------------------------------------------------
; From Delta, set AX to direction 0-3, with 0=left, 1=up, 2=right, 3=down.
;-----------------------------------------------------------------------------
GetDirec:  mov   ax,Delta      ; Row AH -1/1 and column AL 0, or vice versa
           test  ah,ah
           je    iscoldir      ; AH zero?  Then AL -1/1

           xchg  ah,al         ; Else AH -1/1
           inc   al            ; 0/2
iscoldir:  inc   al            ; Column direction 0/2 or row 1/3
           ret
;-----------------------------------------------------------------------------
; Rotate 4x4 block BP clockwise 0-3 times, depending on cursor direction.
;-----------------------------------------------------------------------------
RotBlk:    push  dx
           push  cx
           call  GetDirec      ; Set AX to 0-3
           MOVE  cx,ax
           jcxz  norot         ; Zero?  Done

rcloop:    push  cx            ; Else loop
           xor   ax,ax         ; Clockwise rotate BP to AX by 90 degrees
           mov   cl,16         ; CH zero
rtloop:    xor   dx,dx         ; Clear location for next bit
           shr   bp,1          ; Get next bit
           rcr   dx,cl         ; Rotate through 17-bit positions
           rcr   dx,cl         ;   four times to get final
           rcr   dx,cl         ;   rotated bit location
           rcr   dx,cl
           or    ax,dx         ; Include bit in result AX
           loop  rtloop

           MOVE  bp,ax         ; BP rotated
           pop   cx
           loop  rcloop

norot:     pop   cx
           pop   dx
           ret
;-----------------------------------------------------------------------------
; Kill connected cells at cursor.  Since recursion used, protect low data.
; If history pointer DI non-zero, save history of kill to temp segment.  Kill
; count returned in BP.
;-----------------------------------------------------------------------------
KillChn:   xor   di,di         ; Flag no saves to temp segment
KillDrag:  call  DispWait      ; Preserves BP/DI
           call  SetDXCX       ; Set DX/CX to cursor position, BP/DI ok
KillDXCX:  mov   si,OFFSET Kill; Fixed throughout
           xor   bp,bp         ; Zero kill count
Kill:      cmp   sp,OFFSET WorkBuf+256
           jb    terminate     ; Limit recursion to protect low data

           call  ChkDXCX       ; Insure row/col not -1 or Wid, BP ok
           jnc   terminate     ; Out of range?  Terminate recursion

           call  GetCell       ; Preserves DX/CX/BP/SI/DI
           jnc   terminate     ; Was dead?  Terminate recursion

           test  di,di
           je    nohist        ; No history?  Skip location save

           mov   es,TempSeg    ; Otherwise save kill location
           push  di            ; Save in case at buffer end
           mov   ax,cx
           stosw
           mov   ax,dx
           stosw
           push  cs
           pop   es
           pop   ax            ; Restore start
           xchg  ax,di         ; DI back to pre-stosw state
           test  ax,ax
           je    terminate     ; End of temp segment?  Terminate recursion

           MOVE  di,ax         ; Update pointer
nohist:    inc   bp            ; Update kill count--could get rollover
           call  ClrCell       ; Kill cell and fall, recursing...
;-----------------------------------------------------------------------------
; Call function SI (which must preserve SI/DX/CX at minimum) at each of 8
; locations surrounding center DX/CX, which is restored on exit.  Function
; called by Kill (recursively) and by ChkGlide.
;-----------------------------------------------------------------------------
Op8SI:     dec   cx
           dec   dx
           call  si
           inc   cx
           call  si
           inc   cx
           call  si
           inc   dx
           call  si
           inc   dx
           call  si
           dec   cx
           call  si
           dec   cx
           call  si
           dec   dx
           call  si
           inc   cx            ; Restore DX/CX to center of 3x3 box
terminate: ret
;-----------------------------------------------------------------------------
; Toggle/display alarm indicator.  Assumes alarm flags in low two bits.
;-----------------------------------------------------------------------------
Alarm:     mov   al,Flags
           test  al,INNUMFLAG
           pushf
           inc   ax            ; Anticipate simple bump toggle
           and   al,3          ; AL now 0-3
           popf
           mov   cx,3          ; Prefix cap
           je    nopfx         ; No number prefix?  Ahead

           call  CapNum        ; AL now 0-3
nopfx:     push  ax
           xchg  al,Flags
           and   al,NOT 3
           or    Flags,al      ; Alarms set
           pop   ax            ; Restore 0-3

           mov   cl,5          ; CH zero
           mul   cl
           add   ax,OFFSET AlarmTbl
           MOVE  si,ax
           mov   di,OFFSET AlarmFoot
           rep   movsb         ; Copy alarm indicator
           ret
;-----------------------------------------------------------------------------
; Set DX/CX to cursor row/col.  AX gets old CX.
;-----------------------------------------------------------------------------
SetDXCX:   mov   ax,Top
           add   ax,Row
           MOVE  dx,ax
           mov   ax,Left
           add   ax,Col
           MOVE  cx,ax
           ret
;-----------------------------------------------------------------------------
; Set carry if cursor row/col DX/CX within allowed range.  AX set to Wid also.
;-----------------------------------------------------------------------------
ChkDXCX:   mov   ax,Wid
           cmp   cx,ax
           jnc   cdcout

           cmp   dx,ax
cdcout:    ret
;-----------------------------------------------------------------------------
; Operate on cell at row/col DX/CX, preserving DX/CX/BP/SI/DI.  Slow operation,
; so use only during pause for edits.  Range check included.
;-----------------------------------------------------------------------------
PutCell:   pushf               ; Save carry
           call  ClrCell       ; Clear bit, preserving DX/CX/BP/SI
           popf                ; Restore input carry
           jnc   cdcout        ; Clear?  Done, else set cell

SetCell:   mov   bx,OFFSET SetBit
           jmp   SHORT OpCell
ClrCell:   mov   bx,OFFSET ClrBit
           jmp   SHORT OpCell
GetCell:   mov   bx,OFFSET GetBit
OpCell:    call  ChkDXCX       ; Also sets AX to Wid
           jnc   cdcout        ; Bad coordinates?  Out with clear carry,
                               ;   which is correct for GetBit
           push  di
           push  dx
           push  cx
           mov   di,dx
           call  Div8
           mul   di            ; Down to row
           MOVE  di,ax
           push  cx
           MOVE  ax,cx
           call  Div8
           add   di,ax         ; Right to column byte
           pop   cx
           and   cl,7
           mov   al,80h
           shr   al,cl         ; Mask to bit offset
           mov   ds,GridSeg
           call  bx            ; Operation below...
           pop   cx
           pop   dx
           pop   di
           jmp   SHORT unpout  ; Restore DS and exit

GetBit:    and   al,[di]
           neg   al            ; Set/clear carry if non-zero/zero
           ret
SetBit:    or    [di],al
           ret
ClrBit:    not   al
           and   [di],al
           ret
;-----------------------------------------------------------------------------
; Unpack row of BP/16 words at Grid SI to BP bytes at DI.
;-----------------------------------------------------------------------------
UnpkRow:   mov   ds,GridSeg
           call  DivBP16
           MOVE  cx,ax
unpkloop:  push  cx
           lodsw
           xchg  al,ah
           MOVE  dx,ax
           mov   cl,8          ; CH zero
unpkinlp:  xor   ax,ax
           shl   dx,1
           adc   al,al
           shl   dx,1
           adc   ah,ah
           stosw
           loop  unpkinlp

           pop   cx
           loop  unpkloop

unpout:    push  cs
           pop   ds
           ret                 ; CX zero on exit
;-----------------------------------------------------------------------------
; Return set carry if cell found on perimeter.  BP is Wid on exit.
;-----------------------------------------------------------------------------
PerimChk:  call  GetSize       ; Set AX/BP/BX--also zeroes DX
           lds   si,DWORD PTR GridOff
           push  ax            ; Save grid size
           call  HChk          ; Check first row
           pop   si
           sub   si,bx         ; Start of last row
           call  HChk          ; Check last row
           mov   si,bx         ; Start of second row
           dec   si            ; Back to end of first row
VChk:      mov   cx,bp
           dec   cx
vloop:     mov   ax,[si]       ; Fetch right byte then left byte of next row
           and   ax,8001h      ; Mask out all but edge bits
           or    dx,ax         ;
           add   si,bx         ; Ahead compressed row
           loop  vloop         ; Loop includes a couple bytes from HChk

           neg   dx            ; Sets carry if anything on edge
           jmp   SHORT unpout  ; Restore DS and exit

HChk:      mov   cx,bx
           shr   cx,1          ; Assume Wid is 16-multiple
hloop:     lodsw
           or    dx,ax         ; Include all bits
           loop  hloop

hret:      ret
;-----------------------------------------------------------------------------
; Reduce grid by EDGEINC on each edge, discarding edge debris.
;-----------------------------------------------------------------------------
Reduce:    mov   ax,Wid
           sub   ax,MINWID
           call  Div16
           MOVE  cx,ax         ; Cap on repetitions
           call  PopFunc       ; Careful--pops next address
Reduce1:   call  GetSize       ; Sets AX/BP/BX, AX unused
           cmp   bl,MINWID/8
           jbe   hret          ; At min?  Do nothing

           mov   ax,EDGEINC
           sub   Left,ax       ; Correctable by ChkBoth below
           sub   Top,ax        ; Ditto
           mul   bx            ; Down EDGEINC rows
           add   ax,EDGEINC/8  ; In EDGEINC columns
           MOVE  si,ax         ; Start offset for copy

           sub   bx,2*EDGEINC/8; New width of compressed row
           sub   bp,2*EDGEINC  ; New grid size
           mov   Wid,bp        ; Also adjust Wid

           les   di,DWORD PTR GridOff
           push  es
           pop   ds            ; Synch segments

           mov   cx,bp
rdloop:    push  cx
           mov   cx,bx
           rep   movsb
           add   si,2*EDGEINC/8
           pop   cx
           loop  rdloop

           call  SetESDS       ; Restore segments then fall to ChkBoth so
                               ;   adjustment cleared before PopFunc sees it
;-----------------------------------------------------------------------------
; Force both Col/Row and Left/Top to valid values.  On exit, inequality flags
; either Left or Top adjustment was made (for use in PopFunc to kill reps).
;-----------------------------------------------------------------------------
ChkBoth:   mov   dx,SWID-1     ; Col/Row max
           mov   di,OFFSET Col
           mov   bx,OFFSET Left
           call  ChkCsr2       ; Insure Col/Row in range
           dec   bx            ; BX has dummy role now
           dec   bx
           mov   di,bx
           mov   dx,Wid        ; Wid
           sub   dx,SWID       ; Left/Top max
           xor   bp,bp         ; Ignore Col/Row adjustments in flag
ChkCsr2:   call  ChkCsr1
           inc   di
           inc   di
           inc   bx
           inc   bx
ChkCsr1:   mov   ax,[di]
           cmp   ax,-EDGEINC   ; Not -1, since Reduce/PgUp/... use EDGEINC
           jb    ccup1

           add   [bx],ax
           xor   ax,ax
           inc   bp            ; Flag adjustment
ccup1:     cmp   ax,dx
           jna   ccup2

           add   [bx],ax
           mov   ax,dx
           sub   [bx],ax
           inc   bp            ; Flag adjustment
ccup2:     mov   [di],ax
           test  bp,bp         ; Equality flag out
           ret
;-----------------------------------------------------------------------------
; From counts at start of WorkBuf, construct message afterward, returning
; message length AX and zero CX.
;-----------------------------------------------------------------------------
MakeMsg:   mov   di,OFFSET WorkBuf+512
           mov   si,OFFSET WorkBuf
           xor   bp,bp         ; Array index 1-256
           call  KATail        ; Store CrLf
wbloop:    inc   bp

           push  di
           mov   ax,bp
           mov   di,OFFSET ChainMsg+CHNMSGLEN-4
           call  StoThree      ; Length of chain
           lodsw               ; Fetch array element
           push  ax
           call  StoFiveX      ; Count of chains, zeroing CX
           pop   ax
           pop   di

           test  ax,ax
           je    nomove        ; Zero count?  Skip line entry

           push  si            ; CH zero from StoFiveX
           mov   si,OFFSET ChainMsg
           mov   cl,CHNMSGLEN
           rep   movsb         ; Copy message line, updating DI, zeroing CX
           pop   si

nomove:    cmp   bp,256
           jne   wbloop

           test  ax,ax         ; Test last element again
           je    KATail        ; No large chains?  Skip plus sign

           mov   BYTE PTR [di-3],"+"

KATail:    mov   ax,0A0Dh      ; CrLf
           stosw
           mov   ax,di         ; Not MOVE
           sub   ax,OFFSET WorkBuf+512+4
           je    kaout         ; No lines in buffer?  Leave AX zero

           add   ax,4          ; Allow for 2 CrLfs
kaout:     ret
;-----------------------------------------------------------------------------
; Return compressed grid byte-size in AX, Wid in BP, and Wid/8 in BX.
;-----------------------------------------------------------------------------
GetSize:   mov   ax,Wid
           mov   bp,ax
           call  Div8          ; 8 bits per byte
           mov   bx,ax
           mul   bp            ; Also zeroes DX
gsret:     ret
;-----------------------------------------------------------------------------
; Display macro literal message at SI, updating PlayPtr.
;-----------------------------------------------------------------------------
ShowMsg:   push  si            ; PlayPtr
           call  ClrMsg        ; Clear message display first, zeroing CX
           pop   si
           mov   dx,PANELWID   ; Character cap
           push  si            ; Save start of message
smgloop:   cmp   cx,dx
           ja    smgnext       ; Skip bump if message too long

           inc   cx            ; Bump count--first bump reduced afterward
smgnext:   mov   al,[si]
           test  al,al
           je    smgout        ; Exit without SI bump if end macro

           inc   si
           cmp   al,'"'
           jne   smgloop       ; Exit with SI bump if end quote, else loop

smgout:    dec   cx            ; Undo initial bump
           mov   PlayPtr,si    ; Update pointer for GetMacKey
           pop   si            ; Message start, with length CX (zero ok)
           mov   di,OFFSET MsgBuf
           sub   dx,cx         ; Find slack between cap and length
           shr   dx,1          ; Half difference for indent
           add   di,dx
           rep   movsb         ; Copy message for display or later logging
           jmp   SHORT dispmsg ; Display message
;-----------------------------------------------------------------------------
; Clear screen message area, with alarm line display also.
;-----------------------------------------------------------------------------
ClrMsg:    mov   di,OFFSET MsgBuf
           mov   cx,PANELWID
           mov   al," "
           rep   stosb
dispmsg:   mov   si,OFFSET MsgBuf
           mov   dx,MSGRC
           jmp   SHORT DispPanLn   ; Display panel line, zeroing CX
;-----------------------------------------------------------------------------
; Record macro key AL if recording and not playing.  Also display count.  AL
; should be good data or else 0 (converted to dot).  Called in Edit.
;-----------------------------------------------------------------------------
ChkRcrd:   test  Flags,RCRDFLAG
           je    gsret         ; Not recording?  Out

           test  Flags,PLAYFLAG
           jne   gsret         ; Recording, but playing different macro?  Out

           mov   di,RcrdPtr
           cmp   di,OFFSET RcrdBuf+MACMAX
           jnb   gsret         ; Full buffer?  Exit without saving key

           cmp   al,"W"
           je    dgtlp         ; Write macro key?  Treat as bad key

           test  al,al
           jne   crsave        ; Good key?  Ahead
                               ; For next, RcrdBuf begins with non-digit
dgtlp:     dec   di            ; Loop discards prefix digits to null or W
           mov   al,[di]       ; Fetch previous character
           call  IsDigit       ; Test if digit
           jc    dgtlp         ; Yes?  Loop, discarding, else re-store

crsave:    stosb               ; Store byte and fall, updating pointer...
;-----------------------------------------------------------------------------
; Set pointer and display record status.  Called from ChkRcrd/MacRcrd/MacSave.
;-----------------------------------------------------------------------------
SetRcrd:   mov   RcrdPtr,di    ; Update pointer
           or    Flags,RCRDFLAG; Insure on
           MOVE  ax,di
           sub   ax,OFFSET RcrdBuf+1
           mov   di,OFFSET AlarmFoot+14
           mov   si,OFFSET FullText+3
           cmp   ax,MACMAX-1
           jb    tosto4        ; Under limit?  Show number

           std                 ; Store leftward--undone below
           jmp   SHORT tomove4 ; Show Full
;-----------------------------------------------------------------------------
; Toggle macro recording flag, reset pointer, and display 0000 or Off.
;-----------------------------------------------------------------------------
MacRcrd:   mov   al,Flags
           and   al,RCRDFLAG   ; AL is not digit
           mov   di,OFFSET RcrdBuf
           stosb               ; Start buffer with non-digit for StripMac
           je    SetRcrd       ; Was off?  Reset pointer and display 0000

RcrdOff:   mov   di,OFFSET AlarmFoot+11
           mov   al,NOT RCRDFLAG
FlagOff:   and   Flags,al
           mov   si,OFFSET AlarmTbl
tomove4:   mov   cx,4
           rep   movsb
           cld                 ; In case from above
DispAlarm: mov   si,OFFSET AlarmFoot
           mov   dx,ALARMRC
DispPanLn: mov   cl,PANELWID
           jmp   SHORT DispChars
;-----------------------------------------------------------------------------
; Display wait message centered in left panel of screen.
;-----------------------------------------------------------------------------
DispWait:  mov   si,OFFSET WaitText
DispChr4:  mov   dx,CENTERRC
           mov   cl,4
DispHlpLn: mov   bx,HELPCOLR   ; Page zero, yellow
           jmp   SHORT dbcolr
;-----------------------------------------------------------------------------
; Display CL characters at SI to screen row DH, starting at column DL.
;-----------------------------------------------------------------------------
DispFour:  mov   cl,4
DispChars: mov   bx,FORECOLR   ; Page zero, white
dbcolr:    mov   ch,0
           jcxz  dbout

dbloop:    push  cx
           mov   ah,2
           int   10h           ; Cursor to row/column DH/DL
           lodsb
           mov   cl,1          ; CH zero
           mov   ah,0Ah
           int   10h           ; Write AL at cursor
           inc   dx            ; Bump column
           pop   cx
           loop  dbloop

dbout:     ret                 ; CX also zeroed
;-----------------------------------------------------------------------------
; Display play status.  Called in GetMacKey.
;-----------------------------------------------------------------------------
DispPlay:  mov   di,PlayPtr    ; For play, show characters remaining
           mov   cx,MACMAX
           push  cx
           mov   al,0
           repne scasb         ; Scan for null end flag
           pop   ax
           sub   ax,cx
           dec   ax
           mov   di,OFFSET AlarmFoot+9
tosto4:    call  StoFour
           jmp   SHORT DispAlarm
;-----------------------------------------------------------------------------
; Push macro state, if any macro running and if not end of macro.  Latter
; allows chaining macros without cost of stack push.
;-----------------------------------------------------------------------------
PushMac:   test  Flags,PLAYFLAG
           je    pmout         ; Nothing playing?  Out

           mov   di,PlayPtr
           cmp   BYTE PTR [di],0
           je    pmout         ; End of current macro?  Out

           mov   ax,MStkPtr    ; Stack moves up in memory
           MOVE  di,ax
           mov   si,OFFSET PlayNum
           movsw               ; Copy macro number and pointer
           movsw               ;   to stack
           MOVE  ax,di
           cmp   ax,OFFSET MStkBuf+MSTKSIZE
           jb    pmset         ; Guard memory--no error, just wrap stack

SetMStack: mov   ax,OFFSET MStkBuf
pmset:     mov   MStkPtr,ax    ; Update stack pointer
pmout:     ret
;-----------------------------------------------------------------------------
; Pop and load macro, if any on stack.  Set carry if empty stack/open error.
;-----------------------------------------------------------------------------
PopMac:    mov   ax,MStkPtr
           sub   ax,4
           cmp   ax,OFFSET MStkBuf
           jc    pmout         ; No macro on stack?  Out with set carry

           mov   MStkPtr,ax    ; Update stack pointer
           MOVE  si,ax
           lodsw               ; Fetch macro number
           mov   CurNum,ax
           lodsw               ; Fetch pointer into macro
           jmp   SHORT popplay ; Ahead
;-----------------------------------------------------------------------------
; Load macro, compress, and set running.  On PopMac entry, PlayPtr differs.
;-----------------------------------------------------------------------------
MacPlay:   call  PushMac       ; Save current macro state, if any
FirstMac:  mov   ax,OFFSET PlayBuf
popplay:   mov   PlayPtr,ax    ; Set initial or reentered pointer
           mov   ax,CurNum
           mov   PlayNum,ax    ; Save for later pushes
           call  SetMac        ; Store digits/MAC to name
           call  ClrNum        ; Clear CurNum and flags (in case nested/first)
           call  ReadOpen      ; Handle to BX or set carry pop-exit
           mov   dx,OFFSET PlayBuf
           mov   cx,MACREADMAX ; Maximum bytes to read
           mov   ah,3Fh        ; Flag read for int 21h/Close
           call  IntClose      ; Pass AX through, possibly adjusted
           MOVE  cx,ax         ; Bytes-read (capped at CX, if read error)
           mov   di,dx         ; Start of buffer, with CX as length
           call  StripMac      ; Strip comments/white space
           mov   ax,MACMAX-1   ; Cap for CX
           cmp   cx,ax
           jb    playup        ; CX under cap?  Keep CX

           MOVE  cx,ax         ; Else set CX to smaller AX
playup:    add   di,cx         ; Point past macro to store end flag
           mov   BYTE PTR [di],0
           or    Flags,PLAYFLAG; Also clears carry
           ret
;-----------------------------------------------------------------------------
; Strip macro at DI (length CX) of white space and comments.  CX is reduced.
; From semicolon to end of line is considered comment.  Quote text allows
; allows internal spaces/semi-colons, but not below space.
;-----------------------------------------------------------------------------
StripMac:  push  di
           mov   si,di         ; Synch pointers

endline:   mov   ah,0          ; 1 flags in comment, 2 flags in quote
smloop:    jcxz  stripout      ; Done?  Ahead to tidy up

           lodsb               ; Fetch next character,
           dec   cx            ;   updating count
           cmp   ah,1          ; Test if in comment text
           jne   smup          ; No?  Ahead

           cmp   al,10         ; Test if line feed
           jne   smloop        ; Not EOL?  Just discard
           jmp   SHORT endline ; EOL?  Discard, clearing flags

smup:      cmp   ah,2          ; Test if in quote text
           jne   inmac         ; No?  Then neither quote/comment

           cmp   al,'"'        ; Look for end quote
           je    clrsave       ; Found?  Store, clearing flags

           cmp   al," "        ; Check for space
           jae   savechr       ; Space or higher?  Store character
           jmp   SHORT smloop  ; Lower?  Discard

inmac:     cmp   al," "        ; Now in macro--flags are clear
           jbe   smloop        ; White space?  Discard

           mov   ah,1          ; Anticipate begin comment
           cmp   al,";"        ; Test if found
           je    smloop        ; Yes?  Discard, with flag set

           mov   ah,2          ; Anticipate begin quote
           cmp   al,'"'        ; Test if found
           je    savechr       ; Yes?  Store, with flag set

clrsave:   mov   ah,0          ; Clear flags
savechr:   stosb               ; Store
           jmp   SHORT smloop  ; Loop to get next character

stripout:  pop   si            ; Also PackMac exit
           mov   cx,di
           sub   cx,si         ; Reduced length
           mov   di,si         ; Original DI
           ret
;-----------------------------------------------------------------------------
; Pack macro at DI (length CX).  CX is reduced.  Packing replaces long
; repetitions of cursor/N/T/X/R/J strings with number reps.  Assumes macro
; state from recording (without comments/quotes/?/etc.)
;-----------------------------------------------------------------------------
PackMac:   push  di
           mov   si,di         ; Synch pointers

           mov   al,0          ; AL is previous character for digit check
pmloop:    jcxz  stripout      ; Done?  Back to tidy up

           call  IsDigit       ; Check if previous was digit
           mov   al,[si]       ; Fetch next, leaving SI unchanged
           jc    putchr        ; Previous was digit?  Put character

           push  di
           push  cx
           mov   di,OFFSET Letters
           mov   cx,NUMREPLET
           repne scasb         ; Check if an allowed repeating letter
           pop   cx
           pop   di
           je    chkrep        ; Yes?  Check repetitions

           cmp   al,"a"        ; Check lower case (cursor keys)
           jae   chkrep        ; Yes?  Also check repetitions

putchr:    movsb               ; Else just copy character and loop
           dec   cx
           jmp   SHORT pmloop

chkrep:    push  di            ; Save a moment
           mov   di,si
           repe  scasb         ; Scan for repetitions of current character
           je    noadj         ; End of data?  No adjustment to DI/CX

           dec   di
           inc   cx
noadj:     mov   dx,di
           sub   dx,si         ; DX now number of reps
           mov   si,di         ; SI updated past reps, CX ok too
           pop   di            ; Original position--not updated

           cmp   dx,4
           ja    packsome      ; More than 4 reps?  Room to pack

           push  cx            ; Else 4 or fewer reps, so just store
           mov   cx,dx
           rep   stosb         ; DI now updated
           pop   cx
           jmp   SHORT pmloop  ; Loop

packsome:  push  ax            ; Save AL
           push  cx            ; Save--SI preserved by StoThree
           scasw               ; Bump DI by 2
           MOVE  ax,dx         ; Repetition number to store
           call  StoThree      ; Store three digits leftward
           pop   cx
           pop   ax
           scasw
           scasw               ; Bump DI by 4
           stosb               ; Store repeated character, bumping DI
           jmp   SHORT pmloop
;-----------------------------------------------------------------------------
; Ready filename with digits and extension.  AX and CurNum input.
;-----------------------------------------------------------------------------
SetMac:    mov   ax,"MC"       ; First/third extension letters
           jmp   SHORT setname

SetDat:    mov   ax,"DT"       ; Ditto
setname:   mov   di,OFFSET FileName+10
           std                 ; CLD restored later in StoNum
           stosb               ; C or T
           mov   al,"A"
           stosb
           mov   al,ah
           stosb               ; M or D
           mov   cx,999        ; Max
           call  CapNum        ; Set AX to minimum of CurNum and CX
StoThreeX: dec   di            ; Skip dot, ready for StoNum
StoThree:  mov   cl,3
           jmp   SHORT StoNum  ; Set filename digits, restoring CLD

StoFiveX:  dec   di
StoFive:   mov   cl,5
           jmp   SHORT StoNum
;-----------------------------------------------------------------------------
; Store word AX as integer at offset DI, storing left, CL places.  Dot entry
; is for print routine adjustment of HP_* strings.  CX zeroed.
;-----------------------------------------------------------------------------
StoDot:    shl   ax,1
           shl   ax,1
StoFour:   mov   cl,4
StoNum:    mov   ch,0
           mov   bx,10
           std
dnloop:    xor   dx,dx
           div   bx
           xchg  ax,dx
           add   al,"0"
           stosb
           xchg  ax,dx
           loop  dnloop

           cld
           ret
;-----------------------------------------------------------------------------
; Randomly fill grid, using feedback shift method.
;-----------------------------------------------------------------------------
RandFill:  call  DispWait      ; Zeroes CX too
           mov   cl,100        ; 100% max -- CH zero
           call  CapNum        ; Preserves CX
           test  Flags,INNUMFLAG
           jne   rfup          ; CurNum supplied?  Ahead

           mov   al,20         ; Else use 20% default
rfup:      xchg  ah,al
           test  ax,ax
           je    ClrGrid       ; Zero %?  Clear

           dec   ax
           div   cl            ; Still 100
           push  ax            ; Fill value in AL
           call  GetSize       ; Set AX/BP/BX, ignore BP/BX
           shr   ax,1
           MOVE  cx,ax
           pop   bx            ; Fix fill value in BL and fix SI
           mov   si,OFFSET LoSeed
           les   di,DWORD PTR GridOff
rfloop:    push  cx
           mov   cx,16
rfwdlp:    mov   ax,[si]       ; Fetch current seed
           mov   dx,[si+2]
           push  cx
           mov   cl,8          ; CH zero
rfbitlp:   shl   ax,1          ; Loop produces next random byte DH
           rcl   dx,1
           jnc   rfskip

           xor   ax,0B553h
           xor   dl,76h
rfskip:    loop  rfbitlp

           mov   [si],ax       ; Save updated seed
           mov   [si+2],dx
           cmp   bl,dh         ; Clear carry if DH less equal fill value
           cmc
           rcl   bp,1          ; Accumulate bits to BP
           pop   cx
           loop  rfwdlp

           MOVE  ax,bp
           stosw
           pop   cx
           loop  rfloop

           jmp   SHORT resetes
;-----------------------------------------------------------------------------
; Zero grid and reset parameters.
;-----------------------------------------------------------------------------
ClrGrid:   mov   cx,MAXSIZE
           les   di,DWORD PTR GridOff
           call  ZeroWds
resetes:   push  cs
           pop   es
Reset:     mov   ax,HISTLEN
           mov   di,OFFSET HistBuf
hsinitlp:  stosb               ; Initialize buffer to unlikely
           dec   ax            ;   repetition pattern
           jne   hsinitlp      ; AX zero on exit for next

           mov   CellMax,ax
           mov   di,OFFSET Gen
           stosw               ; Gen
           stosw               ; GldCnt
tohome:    inc   ah
           stosw               ; Delta - down
           mov   ax,Wid
           sub   ax,SWID
           shr   ax,1
           stosw               ; Left/Top
           stosw
           mov   ax,SWID/2
           stosw               ; Col/Row
           stosw
           ret
;-----------------------------------------------------------------------------
; Home cursor.
;-----------------------------------------------------------------------------
Home:     xor   ax,ax
          mov   di,OFFSET Delta
          jmp   SHORT tohome   ; Back, homing cursor
;-----------------------------------------------------------------------------
; Load bit image from LIFExxx.DAT.
;-----------------------------------------------------------------------------
Load:      call  SetDat        ; Store digits/DAT to name
           call  ReadOpen      ; Handle to BX

           mov   cx,MAXSIZE    ; Maximum compressed bytes-to-read
           mov   ah,3Fh        ; Flag read
           call  RdWrClose     ; Ignore error besides open, setting AX
           MOVE  bx,ax         ; Bytes-read (capped at CX, if read error)
           mov   ax,(MINWID-2*EDGEINC)/4
ldwidlp:   add   ax,2*EDGEINC/4; Determine new Grid size from count of
           mov   cx,ax         ;   bytes-read--on irregular read from file
           mul   cx            ;   of improper size, this makes best guess
           shl   ax,1          ; Since squaring quarter Wid, double to
           cmp   ax,bx         ;   get compressed byte size
           MOVE  ax,cx         ; Quarter Wid to AX for next loop
           jb    ldwidlp

           shl   ax,1          ; Double twice to get Wid
           shl   ax,1
           mov   Wid,ax
           jmp   SHORT Reset   ; Reset Left/Top/Row/Col/Gen/Delta/etc.
;-----------------------------------------------------------------------------
; Check alarms.  Beep and set carry if cycle detected or on edge contact.
; Also in such case, exit current macro, but do not kill calling macro.
;-----------------------------------------------------------------------------
ChkAlarms: test  Flags,HISTALARM
           je    skiphist      ; Not flagged?  Ahead

           mov   si,OFFSET HistBuf
           mov   cx,HISTLEN/2  ; Maximum cycle length to test
hsoutlp:   mov   ax,Gen
           inc   ax
           and   al,HISTLEN-1  ; HISTLEN max is 128
           cbw                 ; Point DI to oldest History element
           MOVE  di,ax         ;   in circular buffer
           mov   bp,cx         ; Save current cycle length
           mov   cl,HISTLEN    ; Loop iterations will be one cycle length
           sub   cx,bp         ;   less than total length of buffer
hsinlp:    mov   bx,di         ; Use BX to mask pointers into buffer
           and   bl,HISTLEN-1
           mov   al,[si+bx]    ; Fetch element--SI fixed
           add   bx,bp
           and   bl,HISTLEN-1
           cmp   al,[si+bx]    ; Compare to element one cycle ahead
           jne   hsnext        ; No match?  Then no cycle

           inc   di            ; Else ahead one element
           loop  hsinlp        ;   and loop
                               ; Whole buffer tested?  Then found a cycle
isalarm:   mov   al,0
           mov   di,PlayPtr    ; Force exit at current macro nesting level,
           stosb               ;   but no harm if no macro (PlayPtr in range)
tobeep:    jmp   Beep          ; Also sets carry

hsnext:    mov   cx,bp         ; Restore cycle length to CX
           loop  hsoutlp       ; Decrement cycle length

skiphist:  test  Flags,EDGEALARM
           je    skipedge

           call  PerimChk      ; Returns equality if ok
           jc    isalarm       ; Contact?  Beep

skipedge:  ret                 ; Carry clear from test
;-----------------------------------------------------------------------------
; Save macro, with pack.
;-----------------------------------------------------------------------------
MacSave:   test  Flags,RCRDFLAG
           je    tobeep        ; Not recording?  Out

           mov   di,OFFSET RcrdBuf+1
           mov   cx,RcrdPtr
           sub   cx,di
           call  PackMac       ; Pack
           add   di,cx         ; DI is new pointer
           call  SetRcrd       ; Update/display pointer

           call  SetMac        ; Store digits/MAC to name
           mov   ah,3Ch        ; Flag create
           call  FileOpen      ; Handle to BX, set carry pop-exit on error
           mov   cx,RcrdPtr
           mov   dx,OFFSET RcrdBuf+1
           sub   cx,dx
           jmp   SHORT HelpWrite  ; Write and close
;-----------------------------------------------------------------------------
; Log state to LIFE.LOG.
;-----------------------------------------------------------------------------
LogState:  call  DispWait      ; Zeroes CX
           mov   ax,CurNum
           test  ax,ax         ; Zero is simple log
           je    skipkill      ; Short log?  Ahead

           call  Catalog       ; Catalog chain info to WorkBuf
           call  MakeMsg       ; Construct message, returning size AX, zero CX
skipkill:  MOVE  bp,ax         ; Save for last write
           mov   di,OFFSET MsgBuf
           mov   si,OFFSET EndLogBuf-OFFSET MsgBuf
           mov   dx,OFFSET LogName
           mov   ax,3D02h      ; Read/write, CX zero
           int   21h
           jnc   lsok          ; Ok?  Ahead, else try create

           mov   di,OFFSET LogHdr+1
           mov   si,OFFSET EndLogBuf-OFFSET LogHdr-1
           mov   ah,3Ch        ; CX zero
           int   21h
           jc    Beep          ; Also failed?  Out with beep

lsok:      MOVE  bx,ax
           mov   ax,4202h      ; Seek to EOF
           cwd                 ; AX high bit zero
           int   21h           ; CX still zero

           mov   dx,di
           mov   cx,si
           mov   ah,40h
           int   21h           ; Write panel counts

           mov   dx,OFFSET WorkBuf+512
           mov   cx,bp         ; Zero ok
HelpWrite: mov   ah,40h
IntClose:  int   21h           ; Write catalog info, if any
           jmp   SHORT Close
;-----------------------------------------------------------------------------
; Open for read/write with AX 3D00h or AH 3Ch.  Beep on error and pop-return.
;-----------------------------------------------------------------------------
ReadOpen:  mov   ax,3D00h      ; Flag read open
FileOpen:  push  ax
           call  DispWait      ; Also zeroes CX
           pop   ax
HelpOpen:  mov   dx,OFFSET FileName
           int   21h
           MOVE  bx,ax         ; Handle, if success
           jnc   beepout

popbeep:   pop   ax            ; Pop return address and fall, setting carry...
;-----------------------------------------------------------------------------
; Beep via the 8253 timer chip.  Carry set on exit.
;-----------------------------------------------------------------------------
Beep:      cmp   CmdArg2,"Q"   ; Quiet flag
           je    nobeep

           push  ax
           call  OneTick       ; Wait for tick rollover

           mov   al,TMR2CTRL   ; Flag two frequency bytes
           out   CTRLPORT,al
           mov   al,0          ; Send frequency (776.8 Hz)
           out   TMR2PORT,al
           mov   al,6          ; Shift left for lower octave
           out   TMR2PORT,al

           in    al,SPKRPORT   ; Activate speaker
           or    al,SPKRCTRL
           out   SPKRPORT,al

           call  OneTick       ; Unless key, one full tick

           in    al,SPKRPORT   ; Deactivate speaker
           and   al,NOT SPKRCTRL
           out   SPKRPORT,al
           pop   ax
nobeep:    stc                 ; Beep always flags error
beepout:   ret
;-----------------------------------------------------------------------------
; Save bit image to LIFExxx.DAT.
;-----------------------------------------------------------------------------
Save:      call  SetDat        ; Store digits/DAT to name
           mov   ah,3Ch        ; Flag create
           call  FileOpen      ; Handle to BX
           push  bx
           call  GetSize       ; Set AX/BP/BX, ignore BP/BX
           pop   bx
           MOVE  cx,ax
           mov   ah,40h        ; Flag write
RdWrClose: call  RdWrBuf       ; Ignore error besides open...mite sloppy
;-----------------------------------------------------------------------------
; Close file.  Return minimum of initial AX/CX in AX.  For use if read
; operation precedes close, and want to ignore error and keep AX in range.
;-----------------------------------------------------------------------------
Close:     push  ax
           mov   ah,3Eh        ; Close file
           int   21h
           pop   ax
           cmp   ax,cx
           jbe   clup

           mov   ax,cx
clup:      ret
;-----------------------------------------------------------------------------
; Wait for one BIOS tick change or until keypress.  Return AX low timer word.
;-----------------------------------------------------------------------------
OneTick:   xor   ax,ax
           mov   ds,ax                 ; Timer is in BIOS data area
           mov   ax,Timer
twloop:    push  ax
           mov   ah,1
           int   16h
           pop   ax
           jne   twout                 ; Key waiting?  Out

           cmp   ax,Timer
           je    twloop                ; No change?  Loop

twout:     jmp   SHORT rwup            ; Restore DS and exit
;-----------------------------------------------------------------------------
; Read/write grid with AH as 3Fh/40h and CX as byte count.
;-----------------------------------------------------------------------------
RdWrBuf:   lds   dx,DWORD PTR GridOff
           int   21h
rwup:      push  cs
           pop   ds
           ret
;-----------------------------------------------------------------------------
; BIOS test LPT1, with beep and pop-exit on error.  Test made at start of user
; print/eject request, otherwise assume current request succeeds.
;-----------------------------------------------------------------------------
ChkPrn:    call  DispWait
           mov   ah,2
           cwd                 ; Zero DX--LPT1
           int   17h
           and   ax,2900h      ; Test for paper out, timeout, I/O error
           jne   popbeep       ; Yes?  Beep and pop-exit

prnok:     ret                 ; AX zero...
;-----------------------------------------------------------------------------
; Ready new sheet coordinates, with possible eject.  Input BP is print width
; (usually Wid, but capped at MAXPRN) and BH/CX zero.
;-----------------------------------------------------------------------------
ReadyPrn:  cmp   cx,PrnWid     ; Test if zero--initially and after manual eject
           je    newpage       ; New sheet?  Ahead

           mov   ax,bp         ; Wid
           xchg  PrnWid,ax
           cmp   bp,ax         ; Check if grid size was changed
           jne   newrow        ; Yes?  Force new page row

           add   PrnCol,ax     ; Else bump column start
           add   ax,PrnCol
           cmp   ax,DOTWID     ; Check if would exceed right edge
           jb    prnok         ; No?  Then coordinates ok

           mov   ax,bp         ; Wid
newrow:    add   PrnRow,ax     ; Else bump row start and reset column
newcol:    mov   cx,bp         ; Set starting column so that full row will
           mov   ax,DOTWID     ;   be horizontally centered, if completed
           div   cx
           shr   dx,1          ; Halve remainder to get indent
           mov   PrnCol,dx

           xchg  ax,cx         ; This test will succeed if just ejected...
           add   ax,PrnRow     ; CH zero since DOTWID/Wid < 256
           cmp   ax,DOTLEN     ; Check if would exceed bottom edge
           jb    prnok         ; No?  Then coordinates ok

newpage:   xor   ax,ax         ; Else force new page, bypassing ChkPrn
           call  EjectNoCk     ; CH zero from top or as noted above
           mov   si,OFFSET HP_Init
           mov   cl,INITLEN    ; CH zero from Eject
           call  Print         ; Also zeroes CX/DX for no lines in next

           mov   ax,DOTWID/2-52; With no top margin, need to move down some,
           mov   bl,18         ;   i.e., first 12 or so of DOTLEN unprintable
           call  DrawLines     ; No lines--just center title--BH zero

           mov   si,OFFSET PanelTitle+1
           mov   cl,PANELWID   ; CH still zero
           call  Print

           mov   PrnWid,bp     ; Wid
           mov   ax,22         ; Four dots below title
           mov   PrnRow,ax
           jmp   SHORT newcol  ; Adjust column for centering
;-----------------------------------------------------------------------------
; HP print grid in 75 dpi raster to LPT1.  Eject only on full page or on user
; request.  On grid size change, new row is started.  For large grids, only
; central region is printed due to sheet width limit.
;-----------------------------------------------------------------------------
PrnGrid:   call  ChkPrn        ; Test port with pop-exit on error--CX zeroed
           call  GetSize       ; Sets BP to Wid and BX to Wid/8
           xor   si,si         ; Anticipate small grid print with SI start
           mov   di,bx         ;   offset, DI line bump, BX line bytes,
           mov   ax,bp         ;   and BP print rows
           sub   ax,MAXPRN     ; Check if small grid
           jbe   pgsmall       ; Small?  Ahead, else cap values for
                               ;   central region print
           shr   ax,1          ; AX is row/col indent
           mov   si,ax         ; Save a moment
           mul   bx            ; Product is row indent bytes
           xchg  si,ax         ; Save row indent bytes, fetching col indent
           call  Div8          ; AX now col indent bytes
           add   si,ax         ; Sum is indented start offset for printing
           mov   bp,MAXPRN     ; Reduce print rows (PrnWid)
           mov   bx,MAXPRN/8   ; Reduce line bytes (PrnWid/8)

pgsmall:   push  di            ; Save Wid/8
           push  si            ; Save initial grid offset
           push  bx            ; Save PrnWid/8 (BH zero)
           call  ReadyPrn      ; From BP and zero BH/CX, set PrnWid/PrnRow/
                               ;   PrnCol, possibly ejecting first
           mov   ax,PrnCol     ; Draw right box edge
           mov   bx,PrnRow
           mov   dx,bp         ; PrnWid
           add   ax,dx
           xor   cx,cx
           call  DrawLines

           sub   ax,dx         ; Draw bottom box edge
           xchg  cx,dx
           add   bx,cx
           call  DrawLines

           sub   bx,cx         ; Draw top and left edges
           mov   dx,cx         ; Both PrnWid now
           call  DrawLines

           pop   ax            ; PrnWid/8 saved earlier
           push  bp            ; PrnWid
           mov   bp,ax         ; BP now PrnWid/8 for loop below
           mov   di,OFFSET HP_Xfer+XFERLEN-2
           call  StoThree      ; Store PrnWid as size of raster transfers

           mov   si,OFFSET HP_Rast
           mov   cl,RASTLEN    ; CH zero from StoThree
           call  Print         ; Specify raster parameters (75 dpi, etc.)

           pop   cx            ; PrnWid
           pop   bx            ; Grid offset (zero if PrnWid is Wid)
           pop   di            ; Wid/8
hploop:    push  cx
           mov   si,OFFSET HP_Xfer
           mov   cx,XFERLEN
           call  Print         ; Specify transfer count--preserves BX/BP/DI
           mov   si,bx
           mov   cx,bp         ; PrnWid/8
           mov   ds,GridSeg
           call  Print         ; Print PrnWid/8 raster bytes
           push  cs
           pop   ds
           add   bx,di         ; Advance grid offset by Wid/8 bytes
           pop   cx
           loop  hploop

           mov   ax,PrnRow     ; Ready vertical position down some
           add   ax,6          ;   from upper left
           mov   di,OFFSET HP_Pos+11
           call  StoDot

           mov   si,OFFSET HP_Stop
           mov   cl,STOPLEN    ; CH zero from StoDot
           call  Print         ; Stop raster then position digit print

           mov   si,OFFSET PanelFoot
           mov   cl,5          ; CH zero from Print
           jmp   SHORT Print   ; Print generation at upper left and exit
;-----------------------------------------------------------------------------
; Reset HP and clear PrnWid as at game start.  No eject if empty print buffer.
;-----------------------------------------------------------------------------
Eject:     call  ChkPrn        ; Check port, with internal error exit
EjectNoCk: mov   PrnWid,ax     ; AX zero from ChkPrn or set beforehand
           mov   si,OFFSET HP_Reset
           mov   cl,2          ; CH zero from ChkPrn or within ReadyPrn
;-----------------------------------------------------------------------------
; BIOS print CX bytes from SI to LPT1.
;-----------------------------------------------------------------------------
Print:     lodsb
           mov   ah,0
           cwd                 ; LPT1--title position assumes DX zero
           int   17h           ; Mite sloppy--rely on earlier 29h check
           loop  Print

           ret
;-----------------------------------------------------------------------------
; At raster col/row AX/BX draw horizontal/vertical lines of length CX/DX.
;-----------------------------------------------------------------------------
DrawLines: push  ax
           push  bx
           push  cx
           push  dx

           push  dx
           push  cx
           push  bx
           mov   di,OFFSET HP_Pos+6
           call  StoDot
           pop   ax
           mov   di,OFFSET HP_Pos+11
           call  StoDot
           pop   ax
           mov   di,OFFSET HP_Lines+6
           call  StoDot
           pop   ax
           mov   di,OFFSET HP_Lines+14
           call  StoDot
           mov   si,OFFSET HP_Pos
           mov   cl,POSLEN     ; CH zero from StoDot
           call  Print

           pop   dx
           pop   cx
           pop   bx
           pop   ax
           ret
;-----------------------------------------------------------------------------
; Switch resolution.
;-----------------------------------------------------------------------------
Zoom:      mov   al,Flags
           test  al,INNUMFLAG
           je    xorflag       ; No prefix number?  Just toggle state

           or    al,HIRESFLAG  ; Set flag, anticipating non-zero prefix
           cmp   CurNum,0      ; Test prefix
           jne   okflag        ; Non-zero?  Flag ok

xorflag:   xor   al,HIRESFLAG  ; Toggle flag
okflag:    mov   Flags,al
           ret
;-----------------------------------------------------------------------------
; Set parameters for centered copy of bit grid to WorkBuf.  Assume input AX
; and CX are zero.
;
;   SI = Start offset in grid        BP = Bit length of full row (Wid)
;   DI = Start offset in WorkBuf     BX = Byte length of full row
;   CX = Rows to copy                DX = Byte length to copy in row
;-----------------------------------------------------------------------------
SetParms:  mov   di,OFFSET WorkBuf
           MOVE  si,ax         ; Grid offset--zero
           mov   cl,BITWID/8   ; CH zero
           call  GetSize       ; Set AX/BP/BX, ignore AX
           mov   ax,bp
           sub   ax,BITWID
           ja    spbig         ; Big grid?  Ahead

           neg   ax            ; Make positive
           shr   ax,1          ; Halve sets AX to target rows to indent down
           push  ax            ; Save for similar right indent
           mul   cx
           add   di,ax         ; Move down in image over row rectangle
           mov   dx,bx         ; Number of bytes in row to copy (all)
           pop   ax
           call  Div8          ; Convert right indent to bytes
           add   di,ax         ; Indent right
           mov   cx,bp         ; Number of grid rows to copy (all)
           ret

spbig:     shr   ax,1          ; Halve sets AX to source rows to indent down
           push  ax            ; Save for similar right indent
           mul   bx
           add   si,ax         ; Down indent by rectangle
           mov   dx,cx         ; Number of bytes in row to copy (some)
           pop   ax
           call  Div8          ; Convert right indent to bytes
           add   si,ax         ; Indent right
           mov   cx,BITWID     ; Number of grid rows to copy (some)
           ret
;-----------------------------------------------------------------------------
; Draw bit image of grid centered in WorkBuf (or just center of grid for big
; grids).  Include outer rectangle as applicable.  Need about 24K bytes room.
; Input DI should point to WorkBuf.
;-----------------------------------------------------------------------------
BitDraw:   mov   cx,BITWID/8*BITWID
           call  ZeroWds       ; Clear image buffer, zeroing AX/CX

           call  SetParms      ; Set SI/DI/CX/DX/BX/BP, ignore BP
           mov   ds,GridSeg
           shr   dx,1          ; For word move
gdloop:    push  cx
           push  di
           push  si
           mov   cx,dx
           rep   movsw         ; Copy (partial) grid row to WorkBuf
           pop   si
           pop   di
           add   si,bx         ; Advance to next grid row
           add   di,BITWID/8   ; Advance to next WorkBuf row
           pop   cx
           loop  gdloop

           push  cs
           pop   ds

           mov   si,OFFSET RowList
           mov   cl,4          ; CH zero
dcloop:    push  cx            ; Loop superimposes small zoom box
           lodsw               ;   corners on larger image
           MOVE  dx,ax
           push  dx
           call  WinPos        ; Set DX/CX to bit row/col, BX to CORNERLEN
           call  DrawHorz      ; Preserves SI
           pop   dx
           xchg  dl,dh
           call  WinPos        ; Set DX/CX to bit row/col, BX to CORNERLEN
           call  DrawVert      ; Preserves SI
           pop   cx
           loop  dcloop

           call  PrepBox       ; Draw four grid edges
           call  DrawHorz      ; Top edge of grid
           call  PrepBox
           add   dx,ax         ; Adjust row
           call  DrawHorz      ; Bottom edge
           call  PrepBox
           call  DrawVert      ; Left edge
           call  PrepBox
           add   cx,ax         ; Adjust column then fall to do right edge...
;-----------------------------------------------------------------------------
; DrawVert takes DX/CX/BX/DI/BP input and validates before draw.
;-----------------------------------------------------------------------------
DrawVert:  xchg  cx,dx         ; Swap for check
           call  CoorAdj       ; Adjust CX/BX, setting DI/BP and carry
           xchg  cx,dx         ; Unswap
           jc    dvdout        ; Off-screen?  Out, else fall...
;-----------------------------------------------------------------------------
; Draw BX vertical dots at row/col DX/CX in bit grid at DI with size BP.
; Assume input already validated.
;-----------------------------------------------------------------------------
DrawVDots: call  PrepDraw      ; Advance DI, set AL/CX, change BP to bytes
           jcxz  dvdout        ; Zero dot count?  Out

dvdloop:   or    [di],al       ; Handle lead/tail partial byte
           add   di,bp
           loop  dvdloop

dvdout:    ret
;-----------------------------------------------------------------------------
; Set DX/CX to upper left of grid box, BX to edge length, and AX to BX-1.
;-----------------------------------------------------------------------------
PrepBox:   mov   ax,Wid
           inc   ax            ; Adjust AX to reach other corners
           mov   bx,ax
           inc   bx            ; Set BX to border edge dot count
           mov   dx,BITWID     ; Grid of length Wid is centered in
           sub   dx,ax         ;   BITWID image box
           shr   dx,1          ; Halve for indent
           mov   cx,dx         ; DX/CX now top left corner
           ret
;-----------------------------------------------------------------------------
; Adjust left/right bit column (or top/bottom row) coordinates CX/BX, setting
; carry if off-screen.  Also set carry if row (or column) DX is off-screen.
; Also set BP to grid bit size and DI to grid offset.
;-----------------------------------------------------------------------------
CoorAdj:   mov   di,OFFSET WorkBuf
           mov   bp,BITWID
           add   bx,cx         ; Convert BX to right edge
           dec   bx            ;   for comparisons
           cmp   dx,bp
           jnc   badcoor       ; Row off screen?  Out

           test  bx,bx         ; Also clears carry
           js    badcoor       ; Right edge negative?  Set carry

           cmp   bx,bp
           jb    caup1         ; Under edge limit?  Ahead

           mov   bx,bp         ; Else reduce right edge
           dec   bx
caup1:     test  cx,cx
           jns   caup2         ; Left edge not negative?  Ahead

           xor   cx,cx         ; Else advance to left edge
caup2:     cmp   cx,bp         ; Clears carry if off-screen
           jnc   badcoor

           sub   bx,cx         ; Should clear carry
           inc   bx            ; Restore BX to (adjusted) dot count
           ret

badcoor:   stc
caout:     ret
;-----------------------------------------------------------------------------
; From row/column DX/CX, dot count BX, grid offset DI, and grid length BP,
; advance DI to first byte, and return AL as bit pointer in byte, change
; input bit size BP to bytes, and move dot count to CX.
;-----------------------------------------------------------------------------
PrepDraw:  xchg  dx,bp
           MOVE  ax,dx         ; Size of grid
           call  Div8          ; Divide by 8 to get bytes
           push  ax
           mul   bp            ; BP is row now
           add   di,ax         ; Advance DI to row start
           mov   ax,cx         ; Bit column
           call  Div8          ; Divide by 8 to get bytes
           add   di,ax         ; Indent DI right to first partial byte
           and   cl,7          ; Bits to indent in byte
           mov   al,80h
           shr   al,cl         ; Indent AL to bit position in byte
           mov   cx,bx
           pop   bp            ; Original BP divided by 8
           ret
;-----------------------------------------------------------------------------
; DrawHorz takes DX/CX/BX/DI/BP input and validates before draw.
;-----------------------------------------------------------------------------
DrawHorz:  call  CoorAdj       ; Adjust CX/BX, setting DI/BP and carry
           jc    dhdout        ; Off-screen?  Out, else fall...
;-----------------------------------------------------------------------------
; Draw BX horizontal dots at row/col DX/CX in bit grid at DI with size BP.
; Assume input already validated.
;-----------------------------------------------------------------------------
DrawHDots: call  PrepDraw      ; Advance DI and set AL/CX (ignore BP)
           jcxz  dhdout        ; Zero dot length?  Out

           call  dhdloop       ; Store dots up to first full byte
           jne   dhdout        ; Did not reach next byte?  Out
           jcxz  dhdout        ; Up to next byte, but done?  Out

           inc   di            ; To next byte
           push  cx            ; Save for trailing dots
           MOVE  ax,cx
           call  Div8          ; Divide by 8 for byte stores
           MOVE  cx,ax
           mov   al,0FFh
           rep   stosb         ; Middle bytes--ok if CX was zero
           pop   cx            ; Restore
           and   cx,7          ; CX now 0-7 trailing dots
           je    dhdout        ; Zero?  Out

           mov   al,80h        ; Else store 1-7 trailing dots
dhdloop:   or    [di],al       ; Handle lead/tail partial byte
           shr   al,1
           loopne dhdloop

dhdout:    ret
;-----------------------------------------------------------------------------
; Unpack low-resolution square to 0's and 255's at DI.  Width is SWID bytes.
; Each row still needs to be replicated 8 times when copying to video.  Need
; CH zero input.
;-----------------------------------------------------------------------------
UnPkText:  call  GetSize       ; Set AX/BP/BX, ignore AX/BP
           mov   ax,Top
           mul   bx            ; Packed bytes in row
           MOVE  si,ax         ; Top rectangle bytes to skip
           mov   ax,Left
           push  ax            ; Save for low bits
           call  Div16         ; Cannot just do Div8 and skip next shift
           shl   ax,1          ; Convert to words
           add   si,ax         ; Skip left edge words--SI ready now
           pop   ax            ; Restore

           and   ax,0Fh        ; Low bits to ignore with shift
           mov   dx,ax         ; Ready DX for inner loop shift

           mov   cl,SWID       ; Loop counter and used next--CH zero input
           add   al,cl
           test  al,0Fh        ; If non-zero, must bump high nibble
           je    uptup

           add   al,10h
uptup:     and   al,0F0h
           MOVE  bp,ax         ; Input for UnPkRow
uptloop:   push  cx            ; Save for outer loop
           push  si            ; Ditto
           push  dx            ; Save for post-unpack shift
           push  di            ; Ditto
           call  UnPkRow       ; BX/BP preserved by call--zeroes CX
           pop   di
           pop   dx
           mov   si,di         ; Synch
           add   si,dx         ; Skip over DX bytes--for next, CH zero
           mov   cl,SWID       ; Byte-width of low resolution display
shflp:     lodsb               ; Shift row left, converting
           neg   al            ;   1's to 255's
           stosb               ; Final DI past SWID-byte row
           loop  shflp         ;   is ready for next outer loop

           pop   si
           add   si,bx         ; Advance SI to next grid row
           pop   cx
           loop  uptloop

           ret                 ; CX zero on exit
;-----------------------------------------------------------------------------
; Display color AL pixel at row/column DX/CX.  Insure on display image.
;-----------------------------------------------------------------------------
DispBit:   jnc   dbclr

           mov   bx,BITWID
           cmp   cx,bx         ; 480 plus or negative?
           jnc   dbclr         ; Skip write

           cmp   dx,bx         ; 480 plus or negative?
           jnc   dbclr         ; Skip write

           mov   bh,0          ; Page zero
           mov   ah,0Ch        ; AL color input
           int   10h
dbclr:     ret
;-----------------------------------------------------------------------------
; Display full screen with current generation and footer panel.  Also update
; history buffer.  Buffer is updated more often than necessary, but ok.
;-----------------------------------------------------------------------------
DispGen:   call  ChkBoth       ; Insure Top/Left/Row/Col in range
           call  GetSize       ; Set AX/BP/BX--zeroes DX too for sum
           MOVE  cx,ax         ; Bytes in compressed grid
           shr   cx,1          ; Words now
           mov   bx,OFFSET SumTbl
           lds   si,DWORD PTR GridOff
gsloop:    lodsw
           xlatb es:WaitText   ; Convert to 0 - 8 bit sum, dummy WaitText
           xchg  al,ah
           xlatb es:WaitText   ; Convert to 0 - 8 bit sum, dummy WaitText
           add   al,ah
           cbw                 ; Zero AH
           add   dx,ax         ; Accumulate sum
           loop  gsloop

           MOVE  ax,dx         ; Sum to AX
           push  cs
           pop   ds

           push  ax            ; Save sum for Cells display
           mov   si,OFFSET CellMax
           cmp   [si],ax       ; SI for CellMax/Wid/CurNum/Gen/GldCnt
           jae   maxok

           mov   [si],ax       ; Update CellMax if needed
maxok:     mov   di,OFFSET PanelFoot+25
           lodsw               ; Current CellMax
           call  StoFiveX      ; Store for log output only

           pop   ax
           push  ax            ; Save again for history below
           call  StoFiveX      ; Store Cells in panel

           lodsw               ; Wid
           call  StoThreeX     ; Store Wid in panel

           lodsw               ; CurNum
           call  StoThreeX     ; Store n in panel

           lodsw               ; Gen
           push  ax            ; Save
           call  StoFiveX      ; Store Steps in panel

           dec   di
           lodsw               ; GldCnt
           call  StoThreeX     ; Store glider count in panel

           pop   ax            ; Current Gen

           and   al,HISTLEN-1  ; Circular buffer length is two power
           mov   bl,al         ; BH zero from StoThreeX
           pop   ax            ; Cell count from above
           mov   si,OFFSET HistBuf
           mov   [si+bx],al    ; Save low byte of count

           mov   si,OFFSET PanelFoot
           mov   dx,FOOTRC
           call  DispPanLn     ; Display panel footer line

           call  DispAlarm     ; Display alarm line in panel--zeroes CX

           mov   di,OFFSET WorkBuf
           push  di            ; DI for BitDraw/UnPkText with split pop's

           test  Flags,HIRESFLAG
           je    bytegen       ; Low resolution?  Ahead

bitgen:    call  BitDraw       ; Draw (partial) grid image in WorkBuf
           pop   si
           mov   cx,BITWID
           les   di,DWORD PTR VidOff
bgloop:    push  cx
           mov   cx,SWID/2
           rep   movsw
           add   di,80-SWID
           pop   cx
           loop  bgloop

           push  cs
           pop   es
           ret

bytegen:   call  UnPkText      ; Unpack SWID-square to DI--CX zero input
           pop   si
           mov   cl,SWID       ; CH zero from UnPkText
           les   di,DWORD PTR VidOff
bytoutlp:  push  cx
           mov   cl,8          ; 8 repetitions -- CH zero
bytinlp:   push  cx
           push  si
           mov   cl,SWID/2     ; CH zero
           rep   movsw
           add   di,80-SWID
           pop   si
           pop   cx
           loop  bytinlp

           add   si,SWID
           pop   cx
           loop  bytoutlp

           push  cs
           pop   es            ; Fall...
;-----------------------------------------------------------------------------
; Display low resolution sliding pointers.
;-----------------------------------------------------------------------------
DispPtrs:  mov   ax,Left
           clc                 ; Flag bottom pointer
           call  DispPtr
           mov   ax,Top
           stc                 ; Flag right pointer
DispPtr:   pushf
           mov   bx,Wid
           mov   bp,SWID-1
           sub   bx,bp
           dec   bx            ; BX/BP now divisor/multiplier
           mul   bp
           div   bx            ; AX now 0 to SWID-1
           MOVE  cx,ax
           mov   dx,bp
           mov   bp,0110011001100110b
           mov   al,BITCOLR2
           popf
           jnc   dpout         ; Bottom pointer?  Ahead

           xchg  cx,dx         ; Else swap roles
           mov   bp,0000111111110000b
dpout:     jmp   SHORT DispBlkLo
;-----------------------------------------------------------------------------
; Display low/high resolution cursor, with direction.
;-----------------------------------------------------------------------------
DspGenCsr: call  DispGen
DispCsr:   call  SetDXCX
           call  GetCell       ; Sets carry if at live cell
           mov   al,BITCOLR1
           jnc   dcchk         ; Dead?  Ahead

           mov   al,BITCOLR2   ; Else adjust color
dcchk:     push  ax            ; Save color (split pop's below)
           mov   si,OFFSET Col
           lodsw               ; Col, with SI positioned to Row
           test  Flags,HIRESFLAG
           je    setlocsr      ; Low resolution?  Ahead

sethicsr:  MOVE  dx,ax         ; Col to DL
           lodsw
           mov   dh,al         ; Row
           call  WinPos        ; Set DX/CX
           mov   si,OFFSET BitAdj
           mov   bp,BitCsr     ; Load 4x4 cursor image
           call  PrepBlk
           pop   ax
           jmp   SHORT DispBlk ; Ahead

setlocsr:  MOVE  cx,ax         ; Col
           lodsw
           MOVE  dx,ax         ; Row
           mov   bp,ByteCsr    ; Load 4x4 cursor image
           call  RotBlk        ; Rotate, preserving DX/CX
           pop   ax
;-----------------------------------------------------------------------------
; Display 4x4 block BP centered at low resolution DX/CX with color AL.
;-----------------------------------------------------------------------------
DispBlkLo: shl   cx,1          ; Set CX to col pixel offset,
           shl   cx,1          ;   with increment 2 right
           inc   cx            ;   to center 4x4 block
           shl   cx,1

           shl   dx,1          ; Set DX to row pixel offset,
           shl   dx,1          ;   with increment 2 down
           inc   dx            ;   to center 4x4 block
           shl   dx,1          ; Fall...
;-----------------------------------------------------------------------------
; Display 4x4 block BP at high resolution DX/CX with color AL.  Wrap function
; entry with different SI also called by PutList.
;-----------------------------------------------------------------------------
DispBlk:   mov   si,OFFSET DispBit
DoSI4x4:   mov   bx,4          ; Loop 4 rows
pbrowlp:   push  bx
           push  cx
           mov   bl,4          ; Loop 4 columns
pbcollp:   push  bx
           shl   bp,1          ; Get next bit
           call  si            ; DispBit/PutCell, preserving DX/CX/BP/SI
           inc   cx
           pop   bx
           dec   bx
           jne   pbcollp

           pop   cx
           pop   bx
           inc   dx
           dec   bx
           jne   pbrowlp

           ret
;-----------------------------------------------------------------------------
; From low resolution row/col DH/DL, return high resolution coordinates DX/CX.
; Coordinates may be off-screen.  Also set BX to CORNERLEN.
;-----------------------------------------------------------------------------
WinPos:    mov   ax,BITWID
           sub   ax,Wid        ; Negative if large grid

           sar   ax,1          ; Indent displacement to box if small grid,
           MOVE  bx,ax         ;   else negative displacement

           mov   al,dl
           cbw                 ; DH/DL are bounded by SWID, but allow
           add   ax,Left       ;   small negative or slightly over SWID
           add   ax,bx
           MOVE  cx,ax

           mov   al,dh
           cbw                 ; Ditto
           add   ax,Top
           add   ax,bx
           MOVE  dx,ax
           mov   bx,CORNERLEN
           ret
;-----------------------------------------------------------------------------
; Zero row at DI+BP.
;-----------------------------------------------------------------------------
ZeroRow:   add   di,bp
           mov   cx,bp         ; Bytes
ZeroWds:   shr   cx,1          ; Words
           xor   ax,ax
           rep   stosw
zrout:     ret
;-----------------------------------------------------------------------------
; Handle one or more generation steps.
;-----------------------------------------------------------------------------
NextGen:   mov   cx,GENREPMAX  ; Cap on repetitions
           call  PopFunc       ; Careful--pops next address
NextGen1:  call  OneStep       ; Transform grid once
           call  PerimChk      ; Check for edge contact (quick), setting BP
           jnc   zrout         ; No contact?  Done, else fall (slower)...
;-----------------------------------------------------------------------------
; Check edges for gliders and kill any found.  May skip corners, since always
; looking for middle of 3x3 pattern.  For loop convenience, some corners are
; included.  Input BP should be Wid.
;-----------------------------------------------------------------------------
KillGlide: mov   si,OFFSET KillList-2
           xor   di,di         ; Flag no DX/CX swap
           call  KGBoth
           inc   di            ; Flag DX/CX swap
KGBoth:    mov   dx,1          ; If DI non-zero, DX/CX roles swapped
           call  KGOne         ; Check top row or left col
           mov   dx,bp         ; Wid
           dec   dx
           dec   dx            ; Check bottom row or right col
KGOne:     mov   cx,bp         ; Wid
           lodsw               ; Bump SI by word
kgloop:    call  ChkGlide      ; Preserves DX/CX/SI/BP/DI
           loop  kgloop

           ret
;-----------------------------------------------------------------------------
; If center live, compare 8 bits surrounding DX/CX to [SI] and [SI+1].  If
; either matches, check if larger 13x13 box around DX/CX has exactly 5 live
; cells.  If so, kill 3x3 area around DX/CX.  DX/CX/SI/BP/DI preserved.
;-----------------------------------------------------------------------------
ChkGlide:  call  SwapChk       ; Swap DX/CX, if flagged by DI
           call  GetCell
           jnc   SwapChk       ; Center dead?  No glider contact

           push  bp
           push  si
           mov   si,OFFSET BitToBP
           call  Op8SI         ; Pack 8 surrounding bits to BP low
           pop   si
           MOVE  ax,bp         ; 8 bits to AL
           cmp   al,[si]
           je    chkslack      ; Match to [SI]?  Ahead

           cmp   al,[si+1]
           jne   cgout         ; Neither matches?  Not glider

chkslack:  mov   bp,-5         ; Count cells in small surrounding box
           push  si            ;   to insure clear border
           push  dx            ; Save location
           push  cx
           sub   cx,6          ; Backup upper left corner so that
           sub   dx,6          ;   original DX/CX in middle
           mov   si,13         ; SI is loop counter for rows
cboutlp:   push  si
           push  cx
           mov   si,13         ; SI is loop counter for cols
cbinlp:    call  GetCell       ; Set carry if live cell
           adc   bp,0
           inc   cx
           dec   si
           jne   cbinlp

           pop   cx
           pop   si
           inc   dx
           dec   si
           jne   cboutlp

           pop   cx            ; Restore location
           pop   dx
           pop   si
           test  bp,bp
           jne   cgout         ; Not 5 cells?  Ignore

           inc   GldCnt        ; Bump count of escaped gliders
           push  si
           mov   si,OFFSET ClrCell
           call  Op8SI         ; Clear 8 surrounding bits
           call  si            ; Clear center bit
           pop   si
cgout:     pop   bp
SwapChk:   test  di,di         ; 0/1 flag
           je    noswap        ; Zero?  No swap

           xchg  dx,cx         ; Swap roles
noswap:    ret

BitToBP:   call  GetCell       ; Sets carry if live cell
           rcl   bp,1
           ret
;-----------------------------------------------------------------------------
; Do one generation, using space for 5 rows in WorkBuf.
;-----------------------------------------------------------------------------
OneStep:   inc   Gen           ; Bump counter
           call  GetSize       ; Fix BP during transform routines--registers
                               ;   used exclusively rest of way
           mov   di,OFFSET WorkBuf
           call  ZeroRow       ; Zero for off-grid top boundary
           xor   si,si         ; Grid offset
           call  TripleSum     ; Sum to 3rd row and image to 5th
           xor   si,si
           mov   cx,bp
           dec   cx            ; Less one since one GenRow after loop
ngloop:    push  cx
           call  ShiftUp       ; Rows 2 - 5 shifted to 1 - 4
           push  si
           mov   ax,bp
           call  Div8
           add   si,ax         ; Advance SI by compressed Grid row size
           call  TripleSum     ; Sum to 3rd row and image to 5th
           pop   si
           call  GenRow        ; Generate from 3 sum rows and image in 4th
           mov   si,di         ; DI was advanced by GenRow
           pop   cx
           loop  ngloop

           call  ShiftUp       ; Rows 2 - 5 shifted to 1 - 4
           mov   di,OFFSET WorkBuf
           add   di,bp         ; SI preserved by ShiftUp/ZeroRow
           call  ZeroRow       ; Zero for off-grid bottom boundary, fall...
;-----------------------------------------------------------------------------
; From three rows of triple sums and 4th row holding unpacked image of middle
; row, adjust grid row at SI.  On Pack exit, DI set to advanced input SI.
;-----------------------------------------------------------------------------
GenRow:    push  si            ; Save for Pack
           mov   bx,OFFSET WorkBuf
           mov   si,bx
           add   si,bp
           add   si,bp
           add   si,bp         ; 4th row
           push  si            ; Save for Pack
           mov   di,si         ; Synch up SI and DI
           mov   cx,bp
           shr   cx,1          ; Word loop
grloop:    push  bx
           lodsw
           mov   dx,ax         ; Preserve two cells in DX

           mov   ax,[bx]       ; Handle two cells at a time with
           add   bx,bp         ;   parallel sums
           add   ax,[bx]
           add   bx,bp
           add   ax,[bx]       ; AL/AH now counts of 9-cell blocks
           sub   ax,dx         ; Adjust for center cells
           je    update        ; No neighbors?  Sparse area exit

           test  dl,dl
           jne   lolive        ; Live cell?  Check for continued existence

           cmp   al,3          ; Birth condition is exactly 3 neighbors
           jne   lodie         ; No birth?  Ahead

lolive:    shr   al,1          ; Cell lives if 2 or 3 neighbors
           cmp   al,1
           je    hichk         ; Yes?  Live cell is 1

lodie:     mov   al,0          ; Dead cell is 0

hichk:     test  dh,dh
           jne   hilive        ; Live cell?  Check for continued existence

           cmp   ah,3          ; Birth condition is exactly 3 neighbors
           jne   hidie         ; No birth?  Ahead

hilive:    shr   ah,1          ; Cell lives if 2 or 3 neighbors
           cmp   ah,1
           je    update        ; Yes?  Live cell is 1

hidie:     cbw                 ; Dead cell is 0

update:    stosw               ; Store two cells
           pop   bx
           inc   bx            ; Bump pointer by word
           inc   bx
           loop  grloop

           pop   si
           pop   di            ; Fall to pack fourth row...
;-----------------------------------------------------------------------------
; Pack BP bytes at SI to BP/16 words at Grid DI.  BP preserved.
;-----------------------------------------------------------------------------
PackRow:   mov   es,GridSeg
           call  DivBP16
           MOVE  cx,ax
pkloop:    push  cx
           mov   cl,8          ; CH zero
pkinlp:    lodsw
           shl   dx,1
           add   dl,al
           shl   dx,1
           add   dl,ah
           loop  pkinlp

           MOVE  ax,dx
           xchg  al,ah
           stosw
           pop   cx
           loop  pkloop

pkout:     push  cs
           pop   es
           ret
;-----------------------------------------------------------------------------
; Catalog all chains, counting each size to 256-word array, with last entry
; for chains of 256 or more cells.  Array 1-based.  Grid preserved.
;-----------------------------------------------------------------------------
Catalog:   les   di,DWORD PTR TempOff
           lds   si,DWORD PTR GridOff
           call  FullMove      ; Save grid, since kill is destructive
           mov   di,OFFSET WorkBuf
           mov   ch,2          ; CX was zero from save
           call  ZeroWds       ; Also zeroes CX/AX
koutlp:    cwd                 ; Zero DX
kinlp:     MOVE  di,ax         ; Flag no history (zero)
           call  KillDXCX
           MOVE  ax,bp         ; Final kill count
           test  ax,ax
           je    nochain       ; No chain?  Ignore

           test  ah,ah
           je    smallchn      ; Small chain?  Ahead

           mov   ax,256        ; Else use last element
smallchn:  shl   ax,1          ; Word array, 1-based
           add   ax,OFFSET WorkBuf-2
           xchg  di,ax         ; Zero AX, set DI
           inc   WORD PTR [di] ; Bump array element
nochain:   inc   dx            ; Next row
           cmp   dx,Wid
           jb    kinlp

           inc   cx            ; Next column
           cmp   cx,dx         ; DX is Wid
           jb    koutlp

CopyTemp:  les   di,DWORD PTR GridOff
           lds   si,DWORD PTR TempOff
FullMove:  mov   cx,256*128    ; Restore grid
           rep   movsw
SetESDS:   push  cs            ; Reset ES/DS to CS
           pop   ds
           jmp   SHORT pkout   ; Set ES to CS
;-----------------------------------------------------------------------------
; Expand grid by EDGEINC on each edge, with zero fill.
;-----------------------------------------------------------------------------
Expand:    mov   ax,MAXWID
           sub   ax,Wid
           call  Div16
           MOVE  cx,ax         ; Cap on repetitions
           call  PopFunc       ; Careful--pops next address
Expand1:   call  GetSize       ; Set AX/BP/BX--ignore AX
           cmp   bl,MAXWID/8
           jae   suout         ; At max?  Do nothing

           les   di,DWORD PTR TempOff
           mov   ch,255        ; CL irrelevant--720x720 grid < 255*256 bytes
           call  ZeroWds       ; Zero most of temp segment and CX

           mov   cl,EDGEINC    ; CH zero
           mov   ax,bx         ; Old row size in bytes
           add   al,2*EDGEINC/8; Extend old row size to new
           mul   cx            ; AX now size of top zeroed rectangle
           add   ax,EDGEINC/8  ; Include left edge already zeroed
           MOVE  di,ax         ; Start offset in temp segment for copy

           MOVE  ax,cx         ; EDGEINC
           add   Left,ax       ; Adjust Left/Top/Wid
           add   Top,ax
           shl   ax,1
           add   Wid,ax

           lds   si,DWORD PTR GridOff
           mov   cx,bp         ; Rows to copy
exloop:    push  cx
           mov   cx,bx         ; Compressed bytes per row
           rep   movsb         ; Copy row
           add   di,2*EDGEINC/8; Skip DI over zeroed right/left ends
           pop   cx
           loop  exloop

           call  SetESDS       ; Restore segments before full move
           jmp   SHORT CopyTemp; Copy temp segment to grid and exit
;-----------------------------------------------------------------------------
; Shift 2nd through 5th WorkBuf rows up one row, preserving SI.
;-----------------------------------------------------------------------------
ShiftUp:   push  si
           mov   cx,bp
           mov   di,OFFSET WorkBuf
           mov   si,di
           add   si,bp
           shl   cx,1          ; Double
           rep   movsw         ; Word move effectively doubles again
           pop   si
suout:     ret
;-----------------------------------------------------------------------------
; Replace 3rd row with sum of adjacent triples from SI.  Compressed row is
; unpacked to 5th row so it can later be shifted to 4th row for GenRow.  The
; TripleSum routine stays one row ahead of the GenRow routine.
;-----------------------------------------------------------------------------
TripleSum: mov   di,OFFSET WorkBuf
           add   di,bp
           add   di,bp
           push  di
           add   di,bp
           add   di,bp
           push  di
           call  UnpkRow       ; Unpack to 5th row
           pop   si            ; 5th row
           pop   di            ; 3rd row
           mov   cx,bp
           mov   dl,0          ; Set DL to zero as left boundary
tsloop:    lodsw               ; Fetch next middle/right byte pair
           dec   si            ; Back adjust pointer
           xchg  al,dl         ; Preserve middle byte in DL for next loop
           add   al,dl         ; Sum left and middle bytes
           add   al,ah         ; Add in right byte
           stosb               ; Save sum in 3rd row (bottom of set of 3)
           loop  tsloop

           sub   al,ah
           dec   di
           stosb               ; Adjust last sum for zero right boundary
           ret
;-----------------------------------------------------------------------------
; Accumulator right shifts.
;-----------------------------------------------------------------------------
DivBP16:   mov   ax,bp
Div16:     shr   ax,1
Div8:      shr   ax,1
Div4:      shr   ax,1
           shr   ax,1
           ret
;-----------------------------------------------------------------------------
; Initial macro.  Assumed to precede help panel by program startup.
;-----------------------------------------------------------------------------
InitMac    DB   '"Demo macro"'
           DB   "p3rV7IuKnrrhh"
           DB   "TVTT3n4u5Ih"
           DB   "3Tur3Tllppdd6Ip"
           DB   "10TpV"
           DB   "4Ir5e3T4n"
           DB   "3TluT3uTTld5Tllu5TldTT3dTlu3Thh3lnr"
           DB   "5Tr5Te3d"
           DB   "TlTTd3rTTd3lT"
           DB   "hh3p3d"
           DB   '"Press F1"',"?",0
;-----------------------------------------------------------------------------
; On-screen help panel.  Initial byte in each line combines indent and length.
; Panel status buffers follow, placed here for logging convenience.  When the
; panel is displayed at startup, the code bytes are replaced with spaces so
; that the header appears cleanly in the log file.  MsgBuf contains initial
; error message, which is followed by appended documentation.  Both the
; message and documentation are overwritten.
;-----------------------------------------------------------------------------
PanelTitle DB    0 SHL 5 OR 19, "Game of Life, V1.11"
           DB    0 SHL 5 OR  0
           DB    1 SHL 5 OR 16,  "Full DOC: LIFE ?"
           DB    0 SHL 5 OR  0
           DB    2 SHL 5 OR 10,   "F1=Stop/go"
           DB    2 SHL 5 OR 16,   "nI=Insert object"
           DB    2 SHL 5 OR 15,   "nT=Toggle cells"
           DB    3 SHL 5 OR 12,    "K=Kill chain"
           DB    2 SHL 5 OR 13,   "nD=Drag chain"
           DB    1 SHL 5 OR 14,  "C/H=Clear/home"
           DB    2 SHL 5 OR 14,   "nF=Fill random"
           DB    2 SHL 5 OR 13,   "nN=Next steps"
           DB    0 SHL 5 OR 14, "nS/L=Save/load"
           DB    2 SHL 5 OR  7,   "nZ=Zoom"
           DB    0 SHL 5 OR 18, "nX/R=Expand/reduce"
           DB    1 SHL 5 OR 18,  "P/E=HP print/eject"
           DB    2 SHL 5 OR  8,   "nA=Alarm"
           DB    3 SHL 5 OR 14,    "O=Record macro"
           DB    0 SHL 5 OR 15, "nW/M=Write/play"
           DB    3 SHL 5 OR 12,    "V=Veer right"
           DB    2 SHL 5 OR  7,   "nJ=Jump"
           DB    2 SHL 5 OR 13,   "nG=Log status"
           DB    1 SHL 5 OR  8,  "Esc=Quit"
           DB    0 SHL 5 OR  0
LogHdr     DB    0 SHL 5 OR 19, "---Macro Message---"
           DB    0 SHL 5 OR  0
           DB    0 SHL 5 OR 19, "Alarm Play Rcrd Gld"
           DB    0 SHL 5 OR  0
           DB    0 SHL 5 OR 19, "Steps  n  Wid Cells"
           DB    0 SHL 5 OR  0

           DB    " Max",13,10  ; Max cells are logged but not displayed
MsgBuf     DB    "Need 192K/color VGA$"

AlarmFoot  = MsgBuf    + PANELWID + 2
PanelFoot  = AlarmFoot + PANELWID + 2
EndLogBuf  = PanelFoot + PANELWID + 8

DocText    = $                 ; Documentation appended here...
;-----------------------------------------------------------------------------
; Program buffer area, aside from separate 64K grid segment and separate 64K
; temp segment.  WorkBuf is general purpose heap, with maximum use about 28K.
; During Kill recursion, stack may occupy most of WorkBuf.  The buffers are
; 16-aligned for debugging convenience.
;-----------------------------------------------------------------------------
SLACK      = (OFFSET EndLogBuf - OFFSET Start) MOD 16
SLACK      = (16 - SLACK) MOD 16

MStkBuf    = EndLogBuf + SLACK
HistBuf    = EndLogBuf + SLACK + MSTKSIZE
RcrdBuf    = EndLogBuf + SLACK + MSTKSIZE + HISTLEN
SumTbl     = EndLogBuf + SLACK + MSTKSIZE + HISTLEN + MACMAX
PlayBuf    = EndLogBuf + SLACK + MSTKSIZE + HISTLEN + MACMAX + 256
WorkBuf    = EndLogBuf + SLACK + MSTKSIZE + HISTLEN + MACMAX + 256 + MACMAX

Code_Seg   ENDS
           END Start
